import difflib
import re

from loguru import logger
from sweepai.utils.search_and_replace import Match, find_best_match


def generate_diff(old_code, new_code, **kwargs):
    if old_code == new_code:
        return ""
    stripped_old_code = old_code.strip()
    stripped_new_code = new_code.strip()

    default_kwargs = {"n": 5}
    default_kwargs.update(kwargs)

    diff = difflib.unified_diff(
        stripped_old_code.splitlines(keepends=True),
        stripped_new_code.splitlines(keepends=True),
        **kwargs
    )

    diff_text = "".join(diff)

    return diff_text


def generate_ndiff(old_code, new_code, **kwargs):
    if old_code == new_code:
        return ""
    stripped_old_code = old_code.strip()
    stripped_new_code = new_code.strip()

    diff = difflib.ndiff(
        stripped_old_code.splitlines(keepends=True),
        stripped_new_code.splitlines(keepends=True),
        **kwargs
    )

    diff_text = "".join(diff)

    return diff_text


def revert_whitespace_changes(original_file_str, modified_file_str):
    original_lines = original_file_str.split("\n")
    modified_lines = modified_file_str.split("\n")

    diff = difflib.SequenceMatcher(None, original_lines, modified_lines)

    final_lines = []
    for opcode in diff.get_opcodes():
        if opcode[0] == "equal" or opcode[0] == "replace":
            # If the lines are equal or replace (means the change is not whitespace only)
            # use original lines.
            final_lines.extend(original_lines[opcode[1] : opcode[2]])
        elif opcode[0] == "insert":
            # If the lines are inserted in the modified file, check if it's just whitespace changes
            # If it's just whitespace changes, ignore them.
            for line in modified_lines[opcode[3] : opcode[4]]:
                if line.strip() != "":
                    final_lines.append(line)

    return "\n".join(final_lines)


def format_contents(file_contents, is_markdown=False):
    """
    Add arbitrary postprocessing here, this affects files and diffs
    """
    lines = file_contents.split("\n")

    # Handle small files
    if len(lines) <= 5:
        start_idx = 0
        end_idx = len(lines)
        for idx, line in enumerate(lines):
            if start_idx == 0 and line.strip().startswith("```"):
                start_idx = idx + 1
            if start_idx != 0 and line.strip().endswith("```"):
                end_idx = idx
        lines = lines[start_idx:end_idx]
        return "\n".join(lines)

    first_three_lines = lines[:3]
    last_three_lines = lines[-3:]
    first_line_idx = 0
    last_line_idx = 3
    for idx, line in enumerate(first_three_lines):
        line = line.strip()
        if line.startswith("```"):
            first_line_idx = max(first_line_idx, idx + 1)
        if "user_code>" in line:
            first_line_idx = max(first_line_idx, idx + 1)
    for idx, line in enumerate(last_three_lines):  # Check in reverse
        line = line.strip()
        if line.endswith("```"):
            last_line_idx = min(idx, last_line_idx)
        if "user_code>" in line:
            last_line_idx = min(idx, last_line_idx)
    first_three_lines = first_three_lines[first_line_idx:]
    last_three_lines = last_three_lines[:last_line_idx]

    lines = first_three_lines + lines[3:-3] + last_three_lines
    return "\n".join(lines)


NOT_FOUND = "NOT_FOUND"
IDENTICAL_LINES = "NO MATCHES FOUND"
MULTIPLE_HITS = "MULTIPLE_HITS"
INCOMPLETE_MATCH = "INCOMPLETE_MATCH"


def match_string(original, search, start_index=None, exact_match=False) -> Match:
    pass

    best_match = find_best_match("\n".join(search), "\n".join(original))
    # else:
    #     best_match = Match(index, index + line_matches, score=100)
    logger.print(best_match)
    return best_match


def lstrip_max(s, chars, max_count):
    count = 0
    for char in s:
        if char in chars and count < max_count:
            count += 1
        else:
            break
    return s[count:]


def get_snippet_with_padding(original, best_match, search):
    snippet = original[best_match.start : best_match.end]

    # Fix whitespace
    if search and len(search[0]) - len(search[0].lstrip()) == 0:
        num_whitespace = len(snippet[0]) - len(snippet[0].lstrip())
        if num_whitespace > 0:
            spaces = (
                snippet[0][0] * num_whitespace
            )  # Use first character (tab or space)
        else:
            spaces = ""
        strip = False
    else:  # Do diff between snippet and search
        # Check multiple lines for their whitespace
        min_whitespace = min([len(s) - len(s.lstrip()) for s in search])
        # Rewrite min as for loop
        min_whitespace = None
        character = " "
        for line in search:
            if (
                min_whitespace is None
                or len(line) - len(line.lstrip()) < min_whitespace
            ):
                min_whitespace = len(line) - len(line.lstrip())
                if min_whitespace > 0:
                    character = line[0]
        spaces = character * min_whitespace
        strip = True

    return snippet, spaces, strip


def sliding_window_replacement(
    original: list[str],
    search: list[str],
    replace: list[str],
    search_context_before=None,
    **kwargs,
):
    if search == replace:
        return original, None, None

    best_match = match_string(original, search)
    logger.print(best_match)
    max_similarity = best_match.score

    # No changes could be found. Return original code.
    if max_similarity == 0:
        logger.print("No matches found")
        return original, best_match, NOT_FOUND

    if max_similarity < 50:
        logger.print(f"Low similarity: {max_similarity}")
        return original, best_match, NOT_FOUND

    snippet, spaces, strip = get_snippet_with_padding(original, best_match, search)
    if len(snippet) == 1 and len(replace) == 1:
        # Replace the line
        modified = [snippet[0].replace(search[0], replace[0])]
    elif strip:
        first_line_spaces = min([len(s) - len(s.lstrip()) for s in search])
        modified = [
            spaces
            + (lstrip_max(line, [" ", "\t"], first_line_spaces) if strip else line)
            for line in replace
        ]
    else:
        modified = [spaces + line for line in replace]

    for i in range(min(len(modified) * 2, 40), 0, -1):
        modified_str = "\n".join(modified[:i])
        original_str = "\n".join(original[best_match.start - i : best_match.start])
        modified_pref = "\n".join(modified[-i:])
        original_pref = "\n".join(original[best_match.end : best_match.end + i])
        if modified_str.strip("\n") == original_str.strip("\n"):
            modified = modified[i:]
        if modified_pref == original_pref:
            modified = modified[:-i]
    original = original[: best_match.start] + modified + original[best_match.end :]
    return original, best_match, None


def get_all_diffs(modify_file_response: str) -> str:
    matches = re.findall(
        r"(<<<<.*?\n(.*?)\n====[^\n=]*\n(.*?)\n?>>>>)", modify_file_response, re.DOTALL
    )
    result = "\n\n".join([_match for _match, *_ in matches])
    return result


def get_matches(modify_file_response):
    matches = re.findall(
        r"<<<<.*?\n(.*?)\n====[^\n=]*\n(.*?)\n?>>>>", modify_file_response, re.DOTALL
    )
    return matches


def join_contents_k(first, second, k):
    """
    Join contents together removing k duplicate lines
    """
    first_lines = first.split("\n")
    second_lines = second.split("\n")
    for i in range(k, 0, -1):
        if len(first_lines) < k or len(second_lines) < k:
            continue
        if first_lines[-i:] == second_lines[:i]:
            return "\n".join(first_lines) + "\n" + "\n".join(second_lines[i:])
    return "\n".join(first_lines) + "\n" + "\n".join(second_lines)


def is_markdown(filename):
    return (
        filename.endswith(".md")
        or filename.endswith(".rst")
        or filename.endswith(".txt")
    )


if __name__ == "__main__":
    old_file = """\
            b"sample text",
            b"typed sample text",
            b"3",
        ]
        mock_User_load.return_value = MagicMock()
        mock_TypingPractice.return_value = create_autospec(TypingPractice)

        main(mock_stdscr)

        mock_stdscr.clear.assert_called()
        mock_User_load.assert_called_with(b"username")
        mock_TypingPractice.assert_called()
        mock_TypingPractice.return_value.start_session.assert_called_with(
            b"sample text"
        )
        mock_TypingPractice.return_value.end_session.assert_called_with(
            b"typed sample text"
        )
        mock_TypingPractice.return_value.get_user_statistics.assert_called()
        mock_curses_echo.assert_called()
        mock_curses_initscr.assert_called()

    @patch("touch_typing_practice.main.User.load")
    @patch("touch_typing_practice.main.TypingPractice")
    def test_main_invalid_choice(self, mock_TypingPractice, mock_User_load):
        mock_stdscr = MagicMock()
        mock_stdscr.getstr.side_effect = [
            b"username",
            b"invalid choice",
            b"1",
            b"sample text",
            b"typed sample text",
            b"3",
        ]
        mock_User_load.return_value = MagicMock()
        mock_TypingPractice.return_value = create_autospec(TypingPractice)
        mock_curses_initscr.assert_called()

        main(mock_stdscr)

        mock_stdscr.clear.assert_called()
        mock_User_load.assert_called_with(b"username")
        mock_TypingPractice.assert_called()
        mock_TypingPractice.return_value.start_session.assert_called_with(
            b"sample text"
        )
        mock_TypingPractice.return_value.end_session.assert_called_with(
            b"typed sample text"
        )
        mock_TypingPractice.return_value.get_user_statistics.assert_called()


if __name__ == "__main__":
    unittest.main()"""

    search = """\
        mock_curses_initscr.assert_called()

        main(mock_stdscr)

        mock_stdscr.clear.assert_called()
        mock_User_load.assert_called_with(b"username")
        mock_TypingPractice.assert_called()
        mock_TypingPractice.return_value.start_session.assert_called_with(
            b"sample text"
        )
        mock_TypingPractice.return_value.end_session.assert_called_with(
            b"typed sample text"
        )
        mock_TypingPractice.return_value.get_user_statistics.assert_called()
        mock_curses_echo.assert_called()
        mock_curses_initscr.assert_called()

    @patch("touch_typing_practice.main.User.load")
    @patch("touch_typing_practice.main.TypingPractice")
    def test_main_invalid_choice(self, mock_TypingPractice, mock_User_load):
        mock_stdscr = MagicMock()
        mock_stdscr.getstr.side_effect = [
            b"username",
            b"invalid choice",
            b"1",
            b"sample text",
            b"typed sample text",
            b"3",
        ]
        mock_User_load.return_value = MagicMock()
        mock_TypingPractice.return_value = create_autospec(TypingPractice)
        mock_curses_initscr.assert_called()

        main(mock_stdscr)

        mock_stdscr.clear.assert_called()
        mock_User_load.assert_called_with(b"username")
        mock_TypingPractice.assert_called()
        mock_TypingPractice.return_value.start_session.assert_called_with(
            b"sample text"
        )
        mock_TypingPractice.return_value.end_session.assert_called_with(
            b"typed sample text"
        )
        mock_TypingPractice.return_value.get_user_statistics.assert_called()"""
    replace = """\
        main(mock_stdscr)

        mock_stdscr.clear.assert_called()
        mock_User_load.assert_called_with(b"username")
        mock_TypingPractice.assert_called()
        mock_TypingPractice.return_value.start_session.assert_called_with(
            b"sample text"
        )
        mock_TypingPractice.return_value.end_session.assert_called_with(
            b"typed sample text"
        )
        mock_TypingPractice.return_value.get_user_statistics.assert_called()
        mock_curses_echo.assert_called()

    @patch("touch_typing_practice.main.User.load")
    @patch("touch_typing_practice.main.TypingPractice")
    def test_main_invalid_choice(self, mock_TypingPractice, mock_User_load):
        mock_stdscr = MagicMock()
        mock_stdscr.getstr.side_effect = [
            b"username",
            b"invalid choice",
            b"1",
            b"sample text",
            b"typed sample text",
            b"3",
        ]
        mock_User_load.return_value = MagicMock()
        mock_TypingPractice.return_value = create_autospec(TypingPractice)

        main(mock_stdscr)

        mock_stdscr.clear.assert_called()
        mock_User_load.assert_called_with(b"username")
        mock_TypingPractice.assert_called()
        mock_TypingPractice.return_value.start_session.assert_called_with(
            b"sample text"
        )
        mock_TypingPractice.return_value.end_session.assert_called_with(
            b"typed sample text"
        )
        mock_TypingPractice.return_value.get_user_statistics.assert_called()"""
    print(
        "\n".join(
            sliding_window_replacement(
                old_file.split("\n"), search.split("\n"), replace.split("\n")
            )[0]
        )
    )



from unittest.mock import patch

from sweepai.utils.progress import create_index


@patch("sweepai.utils.progress.global_mongo_client")
def test_create_index(mock_mongo_client):
    mock_db = mock_mongo_client.return_value
    mock_collection = mock_db.__getitem__.return_value

    create_index()

    mock_collection.create_index.assert_called_once_with("tracking_id", unique=True)



import traceback

import resend
from github import Github
from github.AppAuthentication import AppAuthentication
from pydantic import BaseModel

from sweepai.config.server import (
    GITHUB_APP_ID,
    GITHUB_APP_PEM,
    IS_SELF_HOSTED,
    PROGRESS_BASE_URL,
    RESEND_API_KEY,
)
from sweepai.utils.chat_logger import discord_log_error, global_mongo_client
from sweepai.utils.github_utils import get_installation_id

resend.api_key = RESEND_API_KEY


class UserSettings(BaseModel):
    username: str
    email: str = ""
    do_email: bool = True

    @classmethod
    def from_username(
        cls, username: str, installation_id: int = None
    ) -> "UserSettings":
        if IS_SELF_HOSTED:
            return cls(username=username, email="", do_email=False)

        db = global_mongo_client["users"]
        collection = db["users"]

        doc = collection.find_one({"username": username})

        if doc is None:
            # Try get email from github

            try:
                installation_id = get_installation_id(username)
                auth = AppAuthentication(
                    installation_id=installation_id,
                    app_id=GITHUB_APP_ID,
                    private_key=GITHUB_APP_PEM,
                )
                g = Github(app_auth=auth)
                email = (
                    g.get_user(username).email or ""
                )  # Some user's have private emails
            except Exception as e:
                discord_log_error(
                    str(e)
                    + "\n\n"
                    + traceback.format_exc()
                    + f"\n\nUsername: {username}"
                )
                email = ""
            return UserSettings(username=username, email=email)

        return cls(**doc)

    def get_message(self, completed: bool = False) -> str:
        # This is a message displayed to the user in the ticket
        if self.email and self.do_email:
            return f"> [!TIP]\n> I'll email you at {self.email} when I complete this pull request!"
        elif not self.email:
            if not completed:
                return f"> [!TIP]\n> I can email you when I complete this pull request if you set up your email [here]({PROGRESS_BASE_URL}/profile)!"
            else:
                return f"> [!TIP]\n> I can email you next time I complete a pull request if you set up your email [here]({PROGRESS_BASE_URL}/profile)!"

    def send_email(
        self,
        subject: str,
        html: str,
    ):
        if self.email and self.do_email and RESEND_API_KEY is not None:
            return resend.Emails.send(
                {
                    "from": "Sweep Alerts <notifications@sweep.dev>",
                    "to": self.email,
                    "subject": subject,
                    "html": html,
                }
            )


if __name__ == "__main__":
    print(UserSettings.from_username("kevinlu1248").send_email("test", "test"))



from __future__ import annotations

import ast
from io import StringIO
import os
import re
import subprocess
from tempfile import TemporaryDirectory
import tempfile
import traceback
from dataclasses import dataclass
from typing import Optional
import uuid

from pylint.lint import Run
from pylint.reporters.text import TextReporter
import tiktoken
from loguru import logger
from tree_sitter import Node
from tree_sitter_languages import get_parser

from sweepai.core.entities import Snippet
from sweepai.utils.fuzzy_diff import patience_fuzzy_additions


def non_whitespace_len(s: str) -> int:  # new len function
    return len(re.sub("\s", "", s))


def get_line_number(index: int, source_code: str) -> int:
    total_chars = 0
    line_number = 0
    for line_number, line in enumerate(source_code.splitlines(keepends=True), start=1):
        total_chars += len(line)
        if total_chars > index:
            return line_number - 1
    return line_number

@dataclass
class Span:
    # Represents a slice of a string
    start: int = 0
    end: int = 0

    def __post_init__(self):
        # If end is None, set it to start
        if self.end is None:
            self.end = self.start

    def extract(self, s: str) -> str:
        # Grab the corresponding substring of string s by bytes
        return s[self.start : self.end]

    def extract_lines(self, s: str) -> str:
        # Grab the corresponding substring of string s by lines
        return "\n".join(s.splitlines()[self.start : self.end + 1])

    def __add__(self, other: Span | int) -> Span:
        # e.g. Span(1, 2) + Span(2, 4) = Span(1, 4) (concatenation)
        # There are no safety checks: Span(a, b) + Span(c, d) = Span(a, d)
        # and there are no requirements for b = c.
        if isinstance(other, int):
            return Span(self.start + other, self.end + other)
        elif isinstance(other, Span):
            return Span(self.start, other.end)
        else:
            raise NotImplementedError()

    def __len__(self) -> int:
        # i.e. Span(a, b) = b - a
        return self.end - self.start


AVG_CHAR_IN_LINE = 60


def chunk_tree(
    tree,
    source_code: bytes,
    MAX_CHARS=AVG_CHAR_IN_LINE * 200,  # 200 lines of code
    coalesce=AVG_CHAR_IN_LINE * 50,  # 50 lines of code
) -> list[Span]:
    # 1. Recursively form chunks based on the last post (https://docs.sweep.dev/blogs/chunking-2m-files)
    def chunk_node(node: Node) -> list[Span]:
        chunks: list[Span] = []
        current_chunk: Span = Span(node.start_byte, node.start_byte)
        node_children = node.children
        for child in node_children:
            if child.end_byte - child.start_byte > MAX_CHARS:
                chunks.append(current_chunk)
                current_chunk = Span(child.end_byte, child.end_byte)
                chunks.extend(chunk_node(child))
            elif child.end_byte - child.start_byte + len(current_chunk) > MAX_CHARS:
                chunks.append(current_chunk)
                current_chunk = Span(child.start_byte, child.end_byte)
            else:
                current_chunk += Span(child.start_byte, child.end_byte)
        chunks.append(current_chunk)
        return chunks

    chunks = chunk_node(tree.root_node)

    # 2. Filling in the gaps
    if len(chunks) == 0:
        return []
    if len(chunks) < 2:
        end = get_line_number(chunks[0].end, source_code)
        return [Span(0, end)]
    for i in range(len(chunks) - 1):
        chunks[i].end = chunks[i + 1].start
    chunks[-1].end = tree.root_node.end_byte

    # 3. Combining small chunks with bigger ones
    new_chunks = []
    current_chunk = Span(0, 0)
    for chunk in chunks:
        current_chunk += chunk
        # if the current chunk starts with a closing parenthesis, bracket, or brace, we coalesce it with the previous chunk
        stripped_contents = current_chunk.extract(source_code.decode("utf-8")).strip()
        first_char = stripped_contents[0] if stripped_contents else ''
        if first_char in [")", "}", "]"] and new_chunks:
            new_chunks[-1] += chunk
            current_chunk = Span(chunk.end, chunk.end)
        # if the current chunk is too large, create a new chunk, otherwise, combine the chunks
        elif non_whitespace_len(
            current_chunk.extract(source_code.decode("utf-8"))
        ) > coalesce and "\n" in current_chunk.extract(source_code.decode("utf-8")):
            new_chunks.append(current_chunk)
            current_chunk = Span(chunk.end, chunk.end)
    if len(current_chunk) > 0:
        new_chunks.append(current_chunk)

    # 4. Changing line numbers
    first_chunk = new_chunks[0]
    line_chunks = [Span(0, get_line_number(first_chunk.end, source_code))]
    for chunk in new_chunks[1:]:
        start_line = get_line_number(chunk.start, source_code) + 1
        end_line = get_line_number(chunk.end, source_code)
        line_chunks.append(Span(start_line, max(start_line, end_line)))

    # 5. Eliminating empty chunks
    line_chunks = [chunk for chunk in line_chunks if len(chunk) > 0]

    # 6. Coalescing last chunk if it's too small
    if len(line_chunks) > 1 and len(line_chunks[-1]) < coalesce:
        line_chunks[-2] += line_chunks[-1]
        line_chunks.pop()

    return line_chunks


extension_to_language = {
    "js": "tsx",
    "jsx": "tsx",
    "ts": "tsx",
    "tsx": "tsx",
    "mjs": "tsx",
    "py": "python",
    "rs": "rust",
    "go": "go",
    "java": "java",
    "cpp": "cpp",
    "cc": "cpp",
    "cxx": "cpp",
    "c": "cpp",
    "h": "cpp",
    "hpp": "cpp",
    "cs": "cpp",
    "rb": "ruby",
    # "md": "markdown",
    # "rst": "markdown",
    # "txt": "markdown",
    # "erb": "embedded-template",
    # "ejs": "embedded-template",
    # "html": "embedded-template",
    "erb": "html",
    "ejs": "html",
    "html": "html",
    "vue": "html",
    "php": "php",
    "elm": "elm",
}


def naive_chunker(code: str, line_count: int = 30, overlap: int = 15):
    if overlap >= line_count:
        raise ValueError("Overlap should be smaller than line_count.")
    lines = code.split("\n")
    total_lines = len(lines)
    chunks = []

    start = 0
    while start < total_lines:
        end = min(start + line_count, total_lines)
        chunk = "\n".join(lines[start:end])
        chunks.append(chunk)
        start += line_count - overlap

    return chunks

@dataclass
class CheckResults:
    # Experimental feature, we'll see how this does.
    # TODO: smart parsing
    parse_error_message: str = ""
    pylint: str = ""
    eslint: str = ""

    def is_worse_than(self, other: CheckResults) -> bool:
        if self.parse_error_message:
            return True
        if other.parse_error_message:
            return False
        return len(self.pylint.splitlines()) > len(other.pylint.splitlines()) or len(self.eslint.splitlines()) > len(other.eslint.splitlines())
    
    def is_worse_than_message(self, other: CheckResults) -> str:
        if self.parse_error_message:
            return self.parse_error_message
        if other.parse_error_message:
            return other.parse_error_message
        if len(self.pylint.splitlines()) > len(other.pylint.splitlines()):
            # return f"The code has the following pylint errors:\n\n{self.pylint}"
            if not other.pylint:
                return f"The code has the following pylint errors:\n\n{self.pylint}"
            return f"The following new pylint errors have appeared:\n\n{patience_fuzzy_additions(other.pylint, self.pylint)}"
        if len(self.eslint.splitlines()) > len(other.eslint.splitlines()):
            if not other.eslint:
                return f"The code has the following eslint errors:\n\n{self.eslint}"
            return f"The following new eslint errors have appeared:\n\n{patience_fuzzy_additions(other.eslint, self.eslint)}"
        return ""

def strip_ansi_codes(text: str) -> str:
    # ANSI escape sequences (color codes) are often starting with ESC ([) followed by some numbers and ends with "m".
    ansi_escape = re.compile(r'(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]')
    return ansi_escape.sub('', text)

def check_valid_typescript(file_path: str, code: str) -> tuple[bool, str]:
    is_valid = True
    message = ""
    version_check = ["tsc", "--version"]
    result = subprocess.run(
        " ".join(version_check),
        capture_output=True,
        text=True,
        shell=True,
    )
    # only run if tsc is available
    if result.returncode == 0:
        # Create a temporary file to hold the TypeScript code
        with tempfile.NamedTemporaryFile(suffix=".ts", delete=False) as temp_file:
            temp_file_path = temp_file.name
            temp_file.write(code.encode('utf-8'))
        
        # Run `tsc` on the temporary file
        try:
            commands = ["tsc", "--pretty", "--noEmit", temp_file_path]
            result = subprocess.run(" ".join(commands), shell=True, text=True, capture_output=True)

            if result.returncode != 0:
                message = strip_ansi_codes(result.stdout)
                index = message.index(temp_file_path)
                full_temp_file_path = message[:index + len(temp_file_path)]
                message = message.replace(full_temp_file_path, file_path)

                # import error is TS2307 and should come up after the syntax check
                import_error = "error TS2307"
                if import_error in message:
                    num_of_errors = message.count(import_error)
                    # see if this matches the total amount of errors:
                    total_error_message = f"Found {num_of_errors} error"
                    if total_error_message in message:
                        # if we only have import errors, we consider it a successful check
                        return True, ""
                    else:
                        # there are more errors than just import errors
                        # now attempt to parse the message so that we remove import errors
                        message_lines = message.split("\n")
                        while num_of_errors > 0:
                            for line in message_lines:
                                if import_error in line:
                                    message_lines = message_lines[5:]
                                    num_of_errors -= 1
                        message = "\n".join(message_lines)
                return False, message
        finally:
            # Clean up: remove the temporary file
            os.remove(temp_file_path)
    return is_valid, message

def check_syntax(file_path: str, code: str) -> tuple[bool, str]:
    ext = file_path.split(".")[-1]
    if ext in extension_to_language:
        language = extension_to_language[ext]
    else:
        return True, "Unsupported file extension, skipping syntax check."
    parser = get_parser(language)
    tree = parser.parse(code.encode("utf-8"))

    if language == "python":
        # First check for syntax errors
        try:
            ast.parse(code)
        except SyntaxError as e:
            error_message = f"Python syntax error: {e.msg} at line {e.lineno}"
            return False, error_message
    
    # we can't do this right now unfortunately as we don't have a way to mimic the production env for the code
    # if ext in ["ts"]:
    #     return check_valid_typescript(file_path, code)


    def find_deepest_error(node: Node) -> Optional[Node]:
        deepest_error = None
        if node.has_error:
            deepest_error = node
        for child in node.children:
            child_error = find_deepest_error(child)
            if child_error:
                deepest_error = child_error
        return deepest_error

    error_location = find_deepest_error(tree.root_node)
    if error_location:
        start = error_location.start_point
        end = error_location.end_point
        if start[0] == end[0]:
            error_code_lines = [code.split("\n")[start[0]]]
        else:
            error_code_lines = code.split("\n")[start[0]:end[0] + 1]
        error_code_lines[0] = error_code_lines[0][start[1]:]
        error_code_lines[-1] = error_code_lines[-1][:end[1]]
        error_span = "\n".join(error_code_lines)
        if start[0] == end[0]:
            error_message = f"Invalid syntax found at line {start[0]}, displayed below:\n{error_span}"
        else:
            error_message = f"Invalid syntax found from {start}-{end}, displayed below:\n{error_span}"
        return (False, error_message)
    return True, ""

# Need to add "no-unused-vars": "error"
# Need to add "import/first": "error"

DEFAULT_ESLINTRC = """{
  "parser": "@typescript-eslint/parser",
  "parserOptions": {
    "ecmaVersion": 2020,
    "sourceType": "module"
  },
  "plugins": [
    "@typescript-eslint",
    "import"
  ],
  "rules": {
    "no-undef": "error",
    "no-const-assign": "error",
    "no-redeclare": "error",
    "no-unused-vars": "error",
    "no-use-before-define": ["error", { "functions": true, "classes": true, "variables": true }],
    "import/first": "error"
  },
  "settings": {
    "import/resolver": {
      "typescript": {}
    }
  },
  "extends": [
    "eslint:recommended",
    "plugin:@typescript-eslint/recommended",
    "plugin:import/typescript"
  ],
  "overrides": [
    {
      "files": ["*.ts", "*.tsx"],
      "rules": {
        "no-undef": "off"
      }
    }
  ]
}"""

def get_check_results(file_path: str, code: str) -> CheckResults:
    is_valid, error_message = check_syntax(file_path, code)
    if not is_valid:
        return CheckResults(parse_error_message=error_message)
    ext = file_path.split(".")[-1] # noqa
    if ext == "py":
        file_hash = uuid.uuid4().hex
        new_file = os.path.join("/tmp", file_hash + "_" + os.path.basename(file_path))
        stem = os.path.splitext(os.path.basename(file_path))[0]
        try:
            with open(new_file, "w") as f:
                f.write(code)
            pylint_output = StringIO()
            reporter = TextReporter(pylint_output)
            Run(
                [
                    new_file,
                    "--disable=C",
                    "--enable=C0413",  # Enable only the check for imports not at the top
                    "--disable=R",
                    "--disable=import-error",
                    "--disable=no-member",
                    "--disable=unused-import" # we have a workaround for this tbh
                ],
                reporter=reporter,
                exit=False,
            )
            error_message = pylint_output.getvalue().strip()
            try:
                os.remove(new_file)
            except FileNotFoundError:
                pass
            succeeded = error_message.startswith("------------------------------------")
            if error_message:
                error_message = error_message.replace(new_file, file_path).replace(f"{file_hash}_" + stem, stem)
                error_message = error_message.split("-----------------------------------", 1)[0].strip()
                error_message = f"> pylint {file_path}\n\n" + error_message
            return CheckResults(pylint=error_message if not succeeded else "")
        except Exception as e:
            logger.exception(e)
    if ext == "ts":
        # see if eslint is installed
        npx_commands = ["npx", "eslint", "--version"]
        result = subprocess.run(
            " ".join(npx_commands),
            timeout=5,
            capture_output=True,
            text=True,
            shell=True,
        )
        if result.returncode == 0:
            with TemporaryDirectory() as temp_dir:
                new_file = os.path.join(temp_dir, "temp.ts")
                with open(os.path.join(temp_dir, ".eslintrc"), "w") as f:
                    f.write(DEFAULT_ESLINTRC)
                with open(new_file, "w") as f:
                    f.write(code)
                try:
                    eslint_commands = ["npx", "eslint", new_file]
                    result = subprocess.run(
                        " ".join(eslint_commands),
                        capture_output=True,
                        text=True,
                        shell=True,
                        timeout=30,
                    )
                    error_message = (result.stdout + "\n\n" + result.stderr).strip().replace(new_file, file_path)
                    return CheckResults(eslint=error_message)
                except subprocess.TimeoutExpired:
                    logger.warning(f"ESLint timed out after 30s for {file_path}")
                    pass
    return CheckResults()

def check_code(file_path: str, code: str) -> tuple[bool, str]:
    is_valid, error_message = check_syntax(file_path, code)
    if not is_valid:
        return is_valid, error_message
    ext = file_path.split(".")[-1] # noqa
    if ext == "py":
        file_hash = uuid.uuid4().hex
        new_file = os.path.join("/tmp", file_hash + "_" + os.path.basename(file_path))
        stem = os.path.splitext(os.path.basename(file_path))[0]
        try:
            with open(new_file, "w") as f:
                f.write(code)
            pylint_output = StringIO()
            reporter = TextReporter(pylint_output)
            Run(
                [
                    new_file,
                    "--errors-only",
                    "--disable=import-error",
                    "--disable=no-member",
                    "--disable=relative-beyond-top-level",
                ],
                reporter=reporter,
                exit=False,
            )
            error_message = pylint_output.getvalue().strip()
            try:
                os.remove(new_file)
            except FileNotFoundError:
                pass
            if not error_message.startswith("------------------------------------"):
                error_message = error_message.replace(new_file, file_path).replace(f"{file_hash}_" + stem, stem)
                error_message = error_message.split("-----------------------------------", 1)[0].strip()
                error_message = f"> pylint {file_path}\n\n" + error_message
                return False, error_message
        except Exception as e:
            logger.exception(e)
    if ext == "ts":
        # see if eslint is installed
        result = subprocess.run(
            ["npx", "eslint", "--version"],
            capture_output=True,
            text=True,
            timeout=5,
        )
        if result.returncode != 0:
            with TemporaryDirectory() as temp_dir:
                new_file = os.path.join(temp_dir, "temp.ts")
                with open(os.path.join(temp_dir, ".eslintrc"), "w") as f:
                    f.write(DEFAULT_ESLINTRC)
                with open(new_file, "w") as f:
                    f.write(code)
                result = subprocess.run(
                    ["npx", "eslint", new_file, "--config", ".eslintrc"],
                    capture_output=True,
                    text=True,
                    timeout=5,
                )
                if result.returncode != 0:
                    return False, result.stdout + "\n\n" + result.stderr
    return True, ""


# @file_cache()
def chunk_code(
    code: str,
    path: str,
    MAX_CHARS=AVG_CHAR_IN_LINE * 200,  # 200 lines of code
    coalesce=80,
) -> list[Snippet]:
    ext = path.split(".")[-1]
    if ext in extension_to_language:
        language = extension_to_language[ext]
    else:
        # Fallback to naive chunking if tree_sitter fails
        line_count = 50
        overlap = 0
        chunks = naive_chunker(code, line_count, overlap)
        snippets = []
        for idx, chunk in enumerate(chunks):
            end = min((idx + 1) * (line_count - overlap), len(code.split("\n")))
            new_snippet = Snippet(
                content=code,
                start=idx * (line_count - overlap),
                end=end,
                file_path=path,
            )
            snippets.append(new_snippet)
        return snippets
    try:
        parser = get_parser(language)
        tree = parser.parse(code.encode("utf-8"))
        chunks = chunk_tree(
            tree, code.encode("utf-8"), MAX_CHARS=MAX_CHARS, coalesce=coalesce
        )
        snippets = []
        for chunk in chunks:
            new_snippet = Snippet(
                content=code,
                start=chunk.start,
                end=chunk.end,
                file_path=path,
            )
            snippets.append(new_snippet)
        return snippets
    except Exception:
        logger.error(traceback.format_exc())
        return []


TIKTOKEN_CACHE_DIR = "/tmp/cache/tiktoken"


class Tiktoken:
    def __init__(self):
        openai_models = [
            "gpt-3.5-turbo",
            "gpt-3.5-turbo-1106",
            "gpt-4",
            "gpt-4-32k",
            "gpt-4-32k-0613",
            "gpt-4-1106-preview",
            "gpt-4-0125-preview",
        ]
        self.openai_models = {
            model: tiktoken.encoding_for_model(model) for model in openai_models
        }

    def count(self, text: str, model: str = "gpt-4") -> int:
        return len(self.openai_models[model].encode(text, disallowed_special=()))

    def truncate_string(
        self, text: str, model: str = "gpt-4", max_tokens: int = 8192
    ) -> str:
        tokens = self.openai_models[model].encode(text)[:max_tokens]
        return self.openai_models[model].decode(tokens)


test_code = """
import React from 'react';
import { render } from '@testing-library/react';
import CallToAction from '../components/CallToAction';
describe('CallToAction component', () => {
  it('renders the correct YouTube video link', () => {
    const { getByTitle } = render(<CallToAction />);
    const iframeElement = getByTitle('YouTube video player');
    expect(iframeElement.getAttribute('src')).toBe('https://www.youtube.com/embed/GVEkDZmWw8E?autoplay=1&mute=1&loop=1&vq=hd1080&modestbranding=1&controls=0');
    it('has a button with the correct color properties', () => {
    const { getByRole } = render(<CallToAction />);
    const buttonElement = getByRole('button', { name: /install sweep/i });
    expect(buttonElement).toHaveStyle({
      colorScheme: 'green',
      bg: 'green.400',
      _hover: { bg: 'green.600' }
    });
  });
});
"""

test_code = """
x = "test"

import numpy
"""

if __name__ == "__main__":
    # print(check_code("main.tsx", test_code))
    # print(get_check_results("main.py", test_code))
    code = """import { isPossiblyValidEmail } from '../validation-utils'
import { PulseValidationException } from '../pulse-exceptions'

export function getEmailDomain (email: string): string {
  if (!isPossiblyValidEmail(email)) {
    throw new PulseValidationException(`Email is invalid: ${email}`)
  }
  // Emails are tough. An email can contain multiple '@' symbols.
  // Thankfully, domains cannot contain @, so the domain will be
  // part after the last @ in the email.
  // e.g., "steve@macbook"@trilogy.com is a valid email.
  //
  const tokens = email.split('@')
  return tokens[tokens.length - 1].toLowerCase().trim()
}

export function removeEmailAlias(email: string): string {
  if (!isPossiblyValidEmail(email)) {
    throw new PulseValidationException(`Email is invalid: ${email}`)
  
  }
  const atIndex = email.lastIndexOf('@')
  const aliasIndex = email.lastIndexOf('+', atIndex)

  if (aliasIndex > 0) {
    return email.substring(0, aliasIndex) + email.substring(atIndex)
  }

  return email
  }
"""
    new_code = """console.log("hello world")"""
    check_results = check_valid_typescript("test.ts",new_code)
    import pdb
    # pylint: disable=no-member
    pdb.set_trace()



import datetime
import json
import os
import pickle
import threading
import time
import uuid
from itertools import chain, islice

import typer
from github import Github
from github.Event import Event
from github.IssueEvent import IssueEvent
from github.Repository import Repository
from loguru import logger
from rich.console import Console
from rich.prompt import Prompt

from sweepai.api import handle_request
from sweepai.handlers.on_ticket import on_ticket
from sweepai.utils.event_logger import posthog
from sweepai.utils.github_utils import get_github_client
from sweepai.utils.str_utils import get_hash
from sweepai.web.events import Account, Installation, IssueRequest

app = typer.Typer(
    name="sweepai", context_settings={"help_option_names": ["-h", "--help"]}
)
app_dir = typer.get_app_dir("sweepai")
config_path = os.path.join(app_dir, "config.json")

console = Console()
cprint = console.print


def posthog_capture(event_name, properties, *args, **kwargs):
    POSTHOG_DISTINCT_ID = os.environ.get("POSTHOG_DISTINCT_ID")
    if POSTHOG_DISTINCT_ID:
        posthog.capture(POSTHOG_DISTINCT_ID, event_name, properties, *args, **kwargs)


def load_config():
    if os.path.exists(config_path):
        cprint(f"\nLoading configuration from {config_path}", style="yellow")
        with open(config_path, "r") as f:
            config = json.load(f)
        os.environ.update(config)
        os.environ["POSTHOG_DISTINCT_ID"] = str(os.environ.get("POSTHOG_DISTINCT_ID", ""))
        # Should contain:
        # GITHUB_PAT
        # OPENAI_API_KEY
        # ANTHROPIC_API_KEY
        # VOYAGE_API_KEY
        # POSTHOG_DISTINCT_ID


def fetch_issue_request(issue_url: str, __version__: str = "0"):
    (
        protocol_name,
        _,
        _base_url,
        org_name,
        repo_name,
        _issues,
        issue_number,
    ) = issue_url.split("/")
    cprint("Fetching installation ID...")
    installation_id = -1
    cprint("Fetching access token...")
    _token, g = get_github_client(installation_id)
    g: Github = g
    cprint("Fetching repo...")
    issue = g.get_repo(f"{org_name}/{repo_name}").get_issue(int(issue_number))

    issue_request = IssueRequest(
        action="labeled",
        issue=IssueRequest.Issue(
            title=issue.title,
            number=int(issue_number),
            html_url=issue_url,
            user=IssueRequest.Issue.User(
                login=issue.user.login,
                type="User",
            ),
            body=issue.body,
            labels=[
                IssueRequest.Issue.Label(
                    name="sweep",
                ),
            ],
            assignees=None,
            pull_request=None,
        ),
        repository=IssueRequest.Issue.Repository(
            full_name=issue.repository.full_name,
            description=issue.repository.description,
        ),
        assignee=IssueRequest.Issue.Assignee(login=issue.user.login),
        installation=Installation(
            id=installation_id,
            account=Account(
                id=issue.user.id,
                login=issue.user.login,
                type="User",
            ),
        ),
        sender=IssueRequest.Issue.User(
            login=issue.user.login,
            type="User",
        ),
    )
    return issue_request


def pascal_to_snake(name):
    return "".join(["_" + i.lower() if i.isupper() else i for i in name]).lstrip("_")


def get_event_type(event: Event | IssueEvent):
    if isinstance(event, IssueEvent):
        return "issues"
    else:
        return pascal_to_snake(event.type)[: -len("_event")]

@app.command()
def test():
    cprint("Sweep AI is installed correctly and ready to go!", style="yellow")

@app.command()
def watch(
    repo_name: str,
    debug: bool = False,
    record_events: bool = False,
    max_events: int = 30,
):
    if not os.path.exists(config_path):
        cprint(
            f"\nConfiguration not found at {config_path}. Please run [green]'sweep init'[/green] to initialize the CLI.\n",
            style="yellow",
        )
        raise ValueError(
            "Configuration not found, please run 'sweep init' to initialize the CLI."
        )
    posthog_capture(
        "sweep_watch_started",
        {
            "repo": repo_name,
            "debug": debug,
            "record_events": record_events,
            "max_events": max_events,
        },
    )
    GITHUB_PAT = os.environ.get("GITHUB_PAT", None)
    if GITHUB_PAT is None:
        raise ValueError("GITHUB_PAT environment variable must be set")
    g = Github(os.environ["GITHUB_PAT"])
    repo = g.get_repo(repo_name)
    if debug:
        logger.debug("Debug mode enabled")

    def stream_events(repo: Repository, timeout: int = 2, offset: int = 2 * 60):
        processed_event_ids = set()
        current_time = time.time() - offset
        current_time = datetime.datetime.fromtimestamp(current_time)
        local_tz = datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo

        while True:
            events_iterator = chain(
                islice(repo.get_events(), max_events),
                islice(repo.get_issues_events(), max_events),
            )
            for i, event in enumerate(events_iterator):
                if event.id not in processed_event_ids:
                    local_time = event.created_at.replace(
                        tzinfo=datetime.timezone.utc
                    ).astimezone(local_tz)

                    if local_time.timestamp() > current_time.timestamp():
                        yield event
                    else:
                        if debug:
                            logger.debug(
                                f"Skipping event {event.id} because it is in the past (local_time={local_time}, current_time={current_time}, i={i})"
                            )
                if debug:
                    logger.debug(
                        f"Skipping event {event.id} because it is already handled"
                    )
                processed_event_ids.add(event.id)
            time.sleep(timeout)

    def handle_event(event: Event | IssueEvent, do_async: bool = True):
        if isinstance(event, IssueEvent):
            payload = event.raw_data
            payload["action"] = payload["event"]
        else:
            payload = {**event.raw_data, **event.payload}
        payload["sender"] = payload.get("sender", payload["actor"])
        payload["sender"]["type"] = "User"
        payload["pusher"] = payload.get("pusher", payload["actor"])
        payload["pusher"]["name"] = payload["pusher"]["login"]
        payload["pusher"]["type"] = "User"
        payload["after"] = payload.get("after", payload.get("head"))
        payload["repository"] = repo.raw_data
        payload["installation"] = {"id": -1}
        logger.info(str(event) + " " + str(event.created_at))
        if record_events:
            _type = get_event_type(event) if isinstance(event, Event) else "issue"
            pickle.dump(
                event,
                open(
                    "tests/events/"
                    + f"{_type}_{payload.get('action')}_{str(event.id)}.pkl",
                    "wb",
                ),
            )
        if do_async:
            thread = threading.Thread(
                target=handle_request, args=(payload, get_event_type(event))
            )
            thread.start()
            return thread
        else:
            return handle_request(payload, get_event_type(event))

    def main():
        cprint(
            f"\n[bold black on white]  Starting server, listening to events from {repo_name}...  [/bold black on white]\n",
        )
        cprint(
            f"To create a PR, please create an issue at https://github.com/{repo_name}/issues with a title prefixed with 'Sweep:' or label an existing issue with 'sweep'. The events will be logged here, but there may be a brief delay.\n"
        )
        for event in stream_events(repo):
            handle_event(event)

    if __name__ == "__main__":
        main()


@app.command()
def init(override: bool = False):
    # TODO: Fix telemetry
    if not override:
        if os.path.exists(config_path):
            with open(config_path, "r") as f:
                config = json.load(f)
                if "OPENAI_API_KEY" in config and "ANTHROPIC_API_KEY" in config and "GITHUB_PAT" in config:
                    override = typer.confirm(
                        f"\nConfiguration already exists at {config_path}. Override?",
                        default=False,
                        abort=True,
                    )
    cprint(
        "\n[bold black on white]  Initializing Sweep CLI...  [/bold black on white]\n",
    )
    cprint(
        "\nFirstly, let's store your OpenAI API Key. You can get it here: https://platform.openai.com/api-keys\n",
        style="yellow",
    )
    openai_api_key = Prompt.ask("OpenAI API Key", password=True)
    assert len(openai_api_key) > 30, "OpenAI API Key must be of length at least 30."
    assert openai_api_key.startswith("sk-"), "OpenAI API Key must start with 'sk-'."
    cprint(
        "\nNext, let's store your Anthropic API key. You can get it here: https://console.anthropic.com/settings/keys.",
        style="yellow",
    )
    anthropic_api_key = Prompt.ask("Anthropic API Key", password=True)
    assert len(anthropic_api_key) > 30, "Anthropic API Key must be of length at least 30."
    assert anthropic_api_key.startswith("sk-ant-api03-"), "GitHub PAT must start with 'ghp_'."
    cprint(
        "\nGreat! Next, we'll need just your GitHub PAT. Here's a link with all the permissions pre-filled:\nhttps://github.com/settings/tokens/new?description=Sweep%20Self-hosted&scopes=repo,workflow\n",
        style="yellow",
    )
    github_pat = Prompt.ask("GitHub PAT", password=True)
    assert len(github_pat) > 30, "GitHub PAT must be of length at least 30."
    assert github_pat.startswith("ghp_"), "GitHub PAT must start with 'ghp_'."
    cprint(
        "\nAwesome! Lastly, let's get your Voyage AI API key from https://dash.voyageai.com/api-keys. This is optional, but improves code search by about [cyan]3%[/cyan]. You can always return to this later by re-running 'sweep init'.",
        style="yellow",
    )
    voyage_api_key = Prompt.ask("Voyage AI API key", password=True)
    if voyage_api_key:
        assert len(voyage_api_key) > 30, "Voyage AI API key must be of length at least 30."
        assert voyage_api_key.startswith("pa-"), "Voyage API key must start with 'pa-'."

    POSTHOG_DISTINCT_ID = None
    enable_telemetry = typer.confirm(
        "\nEnable usage statistics? This will help us improve the product.",
        default=True,
    )
    if enable_telemetry:
        cprint(
            "\nThank you for enabling telemetry. We'll collect anonymous usage statistics to improve the product. You can disable this at any time by rerunning 'sweep init'.",
            style="yellow",
        )
        POSTHOG_DISTINCT_ID = uuid.getnode()
        posthog.capture(POSTHOG_DISTINCT_ID, "sweep_init", {})

    config = {
        "GITHUB_PAT": github_pat,
        "OPENAI_API_KEY": openai_api_key,
        "ANTHROPIC_API_KEY": anthropic_api_key,
        "VOYAGE_API_KEY": voyage_api_key,
    }
    if POSTHOG_DISTINCT_ID:
        config["POSTHOG_DISTINCT_ID"] = POSTHOG_DISTINCT_ID
    os.makedirs(app_dir, exist_ok=True)
    with open(config_path, "w") as f:
        json.dump(config, f)

    cprint(f"\nConfiguration saved to {config_path}\n", style="yellow")

    cprint(
        "Installation complete! You can now run [green]'sweep run <issue-url>'[/green][yellow] to run Sweep on an issue. or [/yellow][green]'sweep watch <org-name>/<repo-name>'[/green] to have Sweep listen for and fix newly created GitHub issues.",
        style="yellow",
    )


@app.command()
def run(issue_url: str):
    if not os.path.exists(config_path):
        cprint(
            f"\nConfiguration not found at {config_path}. Please run [green]'sweep init'[/green] to initialize the CLI.\n",
            style="yellow",
        )
        raise ValueError(
            "Configuration not found, please run 'sweep init' to initialize the CLI."
        )

    cprint(f"\n  Running Sweep on issue: {issue_url}  \n", style="bold black on white")

    posthog_capture("sweep_run_started", {"issue_url": issue_url})

    request = fetch_issue_request(issue_url)

    try:
        cprint(f'\nRunning Sweep to solve "{request.issue.title}"!\n')
        on_ticket(
            title=request.issue.title,
            summary=request.issue.body,
            issue_number=request.issue.number,
            issue_url=request.issue.html_url,
            username=request.sender.login,
            repo_full_name=request.repository.full_name,
            repo_description=request.repository.description,
            installation_id=request.installation.id,
            comment_id=None,
            edited=False,
            tracking_id=get_hash(),
        )
    except Exception as e:
        posthog_capture("sweep_run_fail", {"issue_url": issue_url, "error": str(e)})
    else:
        posthog_capture("sweep_run_success", {"issue_url": issue_url})


def main():
    cprint(
        "By using the Sweep CLI, you agree to the Sweep AI Terms of Service at https://sweep.dev/tos.pdf",
        style="cyan",
    )
    load_config()
    app()


if __name__ == "__main__":
    main()



import re

from sweepai.core.chat import ChatGPT

prompt = """\
Write a pull request description that reflects all changes in this pull request. Here are the changes:
<diffs>
{diffs}
</diffs>

Here is the pull request title:
<pr_title>
{pr_title}
</pr_title>

Format your response using the following XML tags:
<pr_description>
# Description
Short description of the pull request.
# Summary
Concise bulleted description of the pull request. Markdown format `variables`, `files`, and `directories` like this.
</pr_description>"""


class PRDescriptionBot(ChatGPT):
    def describe_diffs(
        self,
        diffs,
        pr_title,
    ):
        self.messages = []
        # attempt to generate description 3 times
        pr_desc_pattern = r"<pr_description>\n(.*?)\n</pr_description>"
        for attempt in [0, 1, 2]:
            pr_desc_response = self.chat(
                content=prompt.format(
                    diffs=diffs,
                    pr_title=pr_title,
                ),
            )
            pr_desc_matches = re.search(pr_desc_pattern, pr_desc_response, re.DOTALL)
            if pr_desc_matches is None:
                if attempt == 2:
                    return ""
            else:
                break
                
        pr_desc = pr_desc_matches.group(1)
        pr_desc = pr_desc.strip()
        return pr_desc



import time
import traceback

from github.PullRequest import PullRequest
from loguru import logger

from sweepai.agents.pr_description_bot import PRDescriptionBot
from sweepai.config.server import PROGRESS_BASE_URL
from sweepai.core import entities
from sweepai.core.sweep_bot import SweepBot
from sweepai.handlers.create_pr import create_pr_changes
from sweepai.handlers.on_ticket import get_branch_diff_text, sweeping_gif
from sweepai.utils.chat_logger import ChatLogger, discord_log_error
from sweepai.utils.event_logger import posthog
from sweepai.utils.github_utils import ClonedRepo, get_github_client
from sweepai.utils.progress import TicketContext, TicketProgress, TicketProgressStatus
from sweepai.utils.prompt_constructor import HumanMessagePrompt
from sweepai.utils.str_utils import blockquote, to_branch_name
from sweepai.utils.ticket_utils import center, fetch_relevant_files


def stack_pr(
    request: str,
    pr_number: int,
    username: str,
    repo_full_name: str,
    installation_id: int,
    tracking_id: str,
    commit_hash: str = None,
):
    token, g = get_github_client(installation_id=installation_id)
    repo = g.get_repo(repo_full_name)
    pr: PullRequest = repo.get_pull(pr_number)
    branch = pr.head.ref

    if pr.state != "open":
        return {
            "success": False,
            "error_message": "This PR is not open, so I won't attempt to fix it.",
        }

    if (
        sum(
            [
                comment.user.login.startswith("sweep-")
                for comment in pr.get_issue_comments()
            ]
        )
        > 3
    ):
        return {
            "success": False,
            "error_message": "It looks like there are already Sweep comments on this PR, so I won't attempt to fix it.",
        }

    status_message = center(
        f"{sweeping_gif}\n\n"
        + f'Fixing PR: track the progress <a href="{PROGRESS_BASE_URL}/issues/{tracking_id}">here</a>.'
    )
    header = f"{status_message}\n---\n\nI'm currently fixing this PR to address the following:\n\n{blockquote(request)}"
    comment = pr.create_issue_comment(body=header)
    metadata = {}

    def edit_comment(body):
        nonlocal comment
        comment.edit(header + "\n\n" + body)

    try:
        cloned_repo = ClonedRepo(
            repo_full_name=repo_full_name,
            installation_id=installation_id,
            token=token,
            branch=branch,
        )
        start_time = time.time()

        title = request
        if len(title) > 150:
            title = title[:150] + "..."
        ticket_progress = TicketProgress(
            tracking_id=tracking_id,
            username=username,
            context=TicketContext(
                title=title,
                description="",
                repo_full_name=repo_full_name,
                branch_name="sweep/" + to_branch_name(request),
                issue_number=pr_number,
                is_public=repo.private is False,
                start_time=int(time.time()),
            ),
        )

        metadata = {
            "tracking_id": tracking_id,
            "username": username,
            "function": "stack_pr",
            **ticket_progress.context.dict(),
        }
        posthog.capture(
            username,
            "started",
            properties=metadata,
        )

        chat_logger = ChatLogger(
            data={
                "username": username,
                "metadata": metadata,
                "tracking_id": tracking_id,
            }
        )

        is_paying_user = chat_logger.is_paying_user()
        is_consumer_tier = chat_logger.is_consumer_tier()
        issue_url = pr.html_url

        edit_comment("Currently fetching files... (step 1/3)")

        try:
            snippets, tree, _ = fetch_relevant_files(
                cloned_repo,
                request,
                "",
                "",
                username,
                metadata,
                start_time,
                tracking_id,
                is_paying_user,
                is_consumer_tier,
                issue_url,
                chat_logger,
                ticket_progress,
            )
        except Exception as e:
            edit_comment(
                "It looks like an issue has occurred around fetching the files."
                " Perhaps the repo failed to initialized. If this error persists"
                f" contact team@sweep.dev.\n\n> @{username}, editing this issue description to include more details will automatically make me relaunch. Please join our Discord server for support (tracking_id={tracking_id})"
            )
            raise Exception("Failed to fetch files") from e

        ticket_progress.status = TicketProgressStatus.PLANNING
        ticket_progress.save()
        edit_comment("Generating plan by analyzing files... (step 2/3)")

        human_message = HumanMessagePrompt(
            repo_name=repo_full_name,
            issue_url=issue_url,
            username=username,
            repo_description=(repo.description or "").strip(),
            title=request,
            summary=request,
            snippets=snippets,
            tree=tree,
        )

        sweep_bot = SweepBot.from_system_message_content(
            human_message=human_message,
            repo=repo,
            ticket_progress=ticket_progress,
            chat_logger=chat_logger,
            cloned_repo=cloned_repo,
        )
        file_change_requests, plan = sweep_bot.get_files_to_change(
            snippets, cloned_repo
        )

        ticket_progress.status = TicketProgressStatus.CODING
        ticket_progress.save()
        edit_comment("Making changes according to plan... (step 3/3)")
        pull_request = entities.PullRequest(
            title=title,
            branch_name="sweep/" + to_branch_name(request),
            content="",
        )
        generator = create_pr_changes(
            file_change_requests,
            pull_request,
            sweep_bot,
            username,
            installation_id,
            pr_number,
            chat_logger=chat_logger,
            base_branch=branch,
        )

        for item in generator:
            if isinstance(item, dict):
                break
            (
                file_change_request,
                changed_file,
                sandbox_response,
                commit,
                file_change_requests,
            ) = item
            logger.info("Status", file_change_request.status)

        ticket_progress.status = TicketProgressStatus.COMPLETE
        ticket_progress.save()
        edit_comment("Done creating pull request.")

        diff_text = get_branch_diff_text(repo, pull_request.branch_name)
        new_description = PRDescriptionBot().describe_diffs(
            diff_text,
            pull_request.title,
        )
        new_description = (
            f"This pull request was created to fix GitHub Actions on [{commit_hash[:7]}](https://github.com/{repo_full_name}/commit/{commit_hash}).\n\n"
            + new_description
        )
        pull_request.content = new_description
        github_pull_request = repo.create_pull(
            title=pull_request.title,
            body=pull_request.content,
            head=pull_request.branch_name,
            base=branch,
        )

        ticket_progress.context.pr_id = github_pull_request.number
        ticket_progress.context.done_time = time.time()
        ticket_progress.save()
        edit_comment(f"✨ **Created Pull Request:** {github_pull_request.html_url}")
        posthog.capture(
            username,
            "success",
            properties=metadata,
        )
        return {"success": True}
    except Exception as e:
        edit_comment(
            f"> [!CAUTION]\n> \nAn error has occurred: {str(e)} (tracking ID: {tracking_id})"
        )
        discord_log_error(
            traceback.format_exc()
            + "\n\n"
            + str(e)
            + "\n\n"
            + f"tracking ID: {tracking_id}"
        )
        posthog.capture(
            username,
            "failed",
            properties=metadata,
        )
        return {"success": False}


if __name__ == "__main__":
    request = """Fix the GitHub Actions workflow for the repo:

The command:
Run pnpm run build
yielded the following error:
##[error]Process completed with exit code 1.

MongoParseError: Invalid scheme, expected connection string to start with "mongodb://" or "mongodb+srv://"
at new ConnectionString (/home/runner/work/ui/ui/node_modules/.pnpm/mongodb-connection-string-url@3.0.0/node_modules/mongodb-connection-string-url/lib/index.js:86:19)
at parseOptions (/home/runner/work/ui/ui/node_modules/.pnpm/mongodb@6.3.0/node_modules/mongodb/lib/connection_string.js:186:17)
at new MongoClient (/home/runner/work/ui/ui/node_modules/.pnpm/mongodb@6.3.0/node_modules/mongodb/lib/mongo_client.js:51:63)
at 24247 (/home/runner/work/ui/ui/.next/server/app/api/[tracking_id]/route.js:1:1325)
at t (/home/runner/work/ui/ui/.next/server/webpack-runtime.js:1:128)
at 47717 (/home/runner/work/ui/ui/.next/server/app/api/[tracking_id]/route.js:1:471)
at t (/home/runner/work/ui/ui/.next/server/webpack-runtime.js:1:128)
at o (/home/runner/work/ui/ui/.next/server/app/api/[tracking_id]/route.js:1:2997)
at /home/runner/work/ui/ui/.next/server/app/api/[tracking_id]/route.js:1:3024
at t.X (/home/runner/work/ui/ui/.next/server/webpack-runtime.js:1:1206)

> Build error occurred
Error: Failed to collect page data for /api/[tracking_id]
at /home/runner/work/ui/ui/node_modules/.pnpm/next@14.0.3_react-dom@18.0.0_react@18.0.0/node_modules/next/dist/build/utils.js:1217:15
at process.processTicksAndRejections (node:internal/process/task_queues:95:5) {
type: 'Error'
}
ELIFECYCLE  Command failed with exit code 1."""
    stack_pr(
        request=request,
        pr_number=32,
        username="kevinlu1248",
        repo_full_name="sweepai/ui",
        installation_id=36855882,
        tracking_id="fix_typeerror",
    )



import json
from dataclasses import dataclass
from datetime import datetime  # Correctly import the datetime class

import requests
from loguru import logger

from sweepai.config.server import ENV, LOKI_URL, POSTHOG_API_KEY


@dataclass
class PosthogClient:
    """
    Official Posthog API client has a thread leakage, so we are using a custom client.
    """

    API_KEY: str | None = None

    def capture(
        self,
        distinct_id: str | None = None,
        event: str | None = None,
        properties: dict[str, str] | None = None,
        **kwargs,
    ):
        if self.API_KEY is None:
            return
        url = "https://app.posthog.com/capture/"
        headers = {"Content-Type": "application/json"}
        payload = {
            "api_key": self.API_KEY,
            "event": event,
            "properties": {"distinct_id": distinct_id, **properties},
            "timestamp": datetime.utcnow().isoformat()
            + "Z",  # Adding 'Z' to indicate UTC time
        }
        response = requests.post(url, headers=headers, data=json.dumps(payload))
        return response


posthog = PosthogClient(API_KEY=POSTHOG_API_KEY)


def loki_sink(message):
    try:
        record = message.record

        extras = {
            **record["extra"],
            "level": record["level"].name,
            "file": record["file"].path,
            "line": record["line"],
        }

        message = (
            f"{record['time'].isoformat()} {record['level'].name}:{record['file'].path}:{record['line']}: {record['message']}\n\n"
            + json.dumps(extras)
        )

        log_data = {
            "streams": [
                {
                    "stream": {
                        "level": record["level"].name,
                        "env": ENV,
                        # "file": record["file"].path,
                        # "line": record["line"],
                    },
                    "values": [[str(int(record["time"].timestamp() * 1e9)), message]],
                }
            ]
        }

        for key, value in list(record["extra"].items())[:10]:
            if key in ["env"]:
                log_data["streams"][0]["stream"][key] = str(value)

        response = requests.post(
            LOKI_URL,
            data=json.dumps(log_data),
            headers={"Content-Type": "application/json"},
        )
        if response.status_code not in (200, 204):
            print("Error sending log to Loki:", response.text)
    except Exception as e:
        print("Error sending log to Loki:", e)


if LOKI_URL:
    sink_id = logger.add(loki_sink)
    logger.bind(env=ENV)



raise_error_schema = {
    "name": "raise_error",
    "parameters": {
        "type": "object",
        "properties": {
            "message": {
                "type": "string",
                "description": "Message for the user describing the error, either indicating that there's an internal error or that you do not have the necessary information to complete the task. Add all potentially relevant details and use markdown for formatting.",
            }
        },
        "required": ["message"],
    },
    "description": "Use this when you absolutely cannot complete the task on your own.",
}

chain_of_thought_schema = {
    "name": "propose_problem_analysis_and_plan",
    "parameters": {
        "type": "object",
        "properties": {
            "analysis": {
                "type": "string",
                "description": "Break down the problem and identify important pieces of information that will be needed to solve the problem, such as the relevant keywords, the intended behavior, and the required imports.",
            },
            "plan": {
                "type": "string",
                "description": "Describe the plan for the task, including the keywords to search and the modifications to make. Be sure to consider all imports that are required to complete the task.",
            },
        },
        "required": ["analysis", "plan"],
    },
}

search_and_replace_schema = {
    "name": "search_and_replace",
    "parameters": {
        "type": "object",
        "properties": {
            "analysis_and_identification": {
                "type": "string",
                "description": "Identify and list the minimal changes that need to be made to the file, by listing all locations that should receive these changes and the changes to be made. Be sure to consider all imports that are required to complete the task.",
            },
            "replaces_to_make": {
                "type": "array",
                "description": "Array of sections of code to modify.",
                "items": {
                    "type": "object",
                    "properties": {
                        "section_id": {
                            "type": "string",
                            "description": "The section ID the original code belongs to.",
                        },
                        "old_code": {
                            "type": "string",
                            "description": "The old lines of code. Be sure to add lines before and after to disambiguate the change.",
                        },
                        "new_code": {
                            "type": "string",
                            "description": "The new code to replace the old code.",
                        },
                    },
                    "required": ["section_id", "old_code", "new_code"],
                },
            },
        },
        "required": ["analysis_and_identification", "replaces_to_make"],
    },
    "description": "Make edits to the code file.",
}

view_sections_schema = {
    "name": "view_sections",
    "parameters": {
        "type": "object",
        "properties": {
            "section_ids": {
                "type": "array",
                "description": "Section IDs to view",
                "items": {
                    "type": "string",
                    "description": "The section ID to view.",
                },
            },
        },
        "required": ["section_ids"],
    },
    "description": "Searches for sections in the file and returns the code for each section.",
}

keyword_search_schema = {
    "name": "keyword_search",
    "parameters": {
        "type": "object",
        "properties": {
            "justification": {
                "type": "string",
                "description": "Justification for searching the keyword.",
            },
            "keyword": {
                "type": "string",
                "description": "The keyword to search for. This is the keyword itself that you want to search for in the contents of the file, not the name of the file itself.",
            },
        },
        "required": ["justification", "keyword"],
    },
    "description": "Searches for all lines in the file containing the keyword.",
}

submit_schema = {
    "name": "submit",
    "parameters": {
        "type": "object",
        "properties": {
            "justification": {
                "type": "string",
                "description": "Justification for why you are finished with the task.",
            },
        },
        "required": ["justification"],
    },
    "description": "Indicates that you have completed the task successfully.",
}


"""
List of common prompts used across the codebase.
"""

# Following two should be fused
system_message_prompt = """\
You are a brilliant and meticulous engineer assigned to write code for the following Github issue. When you write code, the code works on the first try, is syntactically perfect and is fully implemented. You have the utmost care for the code that you write, so you do not make mistakes and every function and class will be fully implemented. When writing tests, you will ensure the tests are fully implemented, very extensive and cover all cases, and you will make up test data as needed. Take into account the current repository's language, frameworks, and dependencies."""

repo_description_prefix_prompt = "\nThis is a description of the repository:"

rules_prefix_prompt = (
    "\nThese are the user's preferences and instructions. Use them as needed"
)

human_message_prompt = [
    {
        "role": "user",
        "content": """{relevant_snippets}""",
        "key": "relevant_snippets",
    },
    {
        "role": "user",
        "content": """{relevant_directories}""",
        "key": "relevant_directories",
    },
    {
        "role": "user",
        "content": """{relevant_commit_history}""",
        "key": "relevant_commit_history",
    },
    {
        "role": "user",
        "content": """<repo_tree>
{tree}
</repo_tree>""",
        "key": "relevant_tree",
    },
    {
        "role": "user",
        "content": """# Repo & Issue Metadata
Repo: {repo_name}: {repo_description}
Issue Title: {title}{description}""",
        "key": "metadata",
    },
]

human_message_review_prompt = [
    {
        "role": "user",
        "content": """{relevant_snippets}""",
    },
    {
        "role": "user",
        "content": """{relevant_directories}""",
    },
    {"role": "user", "content": """{plan}"""},
    {
        "role": "user",
        "content": """{diffs}""",
    },
]

snippet_replacement_system_message = f"""{system_message_prompt}

You are selecting relevant snippets for this issue. You must only select files that would help you understand the context of this issue.

## Snippet Step

In order to address this issue, what required information do you need about the snippets? Only include relevant code and required file imports that provides you enough detail about the snippets for the problems:

Note: Do not select the entire file. Only select relevant lines from these files. Keep the relevant_snippets as small as possible.

<contextual_thoughts>
* Thought_1
* Thought_2
...
</contextual_thoughts>

<relevant_snippets>
folder_1/file_1.py:1-13
folder_2/file_2.py:42-75
...
</relevant_snippets>
"""

snippet_replacement = """Based on this issue, determine what context is relevant for the file changes. In the relevant_snippets, do not write the entire file lines. Choose only the most important lines.

Complete the Snippet Step."""

diff_section_prompt = """
<file_diff file="{diff_file_path}">
{diffs}
</file_diff>"""

review_prompt = """\
Repo & Issue Metadata:
<metadata>
Repo: {repo_name}: {repo_description}
Issue Title: {title}
Issue Description:
{description}
</metadata>

The code was written by an inexperienced programmer. Carefully review the code diffs in this pull request. Use the diffs along with the original plan to verify that each step of the plan was implemented correctly.

Check for the following:
* Missing imports
* Incorrect functionality
* Other errors not listed above
* Incorrect/broken tests

Indicate all breaking changes. Do not point out stylistic issues. Ensure that the code resolves the issue requested by the user and every function and class is fully implemented.

Respond in the following format:c
<diff_analysis>
Check each file_diff function by function and confirm whether it was both implemented and implemented correctly.
...
</diff_analysis>"""

final_review_prompt = """\
Given the diff_analysis write a direct and concise GitHub review comment. Be extra careful with unimplemented sections and do not nitpick on formatting.
If there is additional work to be done before this PR is ready, mention it. If there are no changes required, simply say "No changes required."
In case changes are required, keep in mind the author is an inexperienced programmer and may need a pointer to the files and specific changes.
Follow this format:
<changes_required>
Write Yes if the changes are required or No if they are not required.
</changes_required>
<review_comment>
Mention any changes that need to be made, using GitHub markdown to format the comment.
- Change required in file on line x1-x2
- Change required in file on line y1-y2
...
</review_comment>"""

issue_comment_prompt = """
<comment username="{username}">
{reply}
</comment>"""

# Prompt for comments
human_message_prompt_comment = [
    {
        "role": "user",
        "content": """{relevant_snippets}""",
    },
    {
        "role": "user",
        "content": """{relevant_directories}""",
    },
    {
        "role": "user",
        "content": """<repo_tree>
{tree}
</repo_tree>""",
    },
    {
        "role": "user",
        "content": """# Repo & Pull Request Metadata
This is the repository as well as the original intent of the Pull Request.
Repo: {repo_name}: {repo_description}
Pull Request Title: {title}
Pull Request Description: {description}{relevant_docs}""",
    },
    {
        "role": "user",
        "content": """These are the previous file changes
{diff}""",
    },
    {
        "role": "user",
        "content": """Please handle the user review comment using the snippets, pull request title, pull request description, and the file changes.
User pull request review: \"{comment}\"""",
    },
]

cot_retrieval_prompt = """
Gather information to solve the problem. Use "finish" when you feel like you have sufficient information.
"""

files_to_change_abstract_prompt = """Write an abstract minimum plan to address this issue in the least amount of change possible. Try to originate the root causes of this issue. Be clear and concise. 1 paragraph."""

files_to_change_system_prompt = """You are a brilliant and meticulous engineer assigned to plan code changes to solve the following Github issue using the snippets provided from the codebase. You have the utmost care for the plan that you write, so you ensure that all the relevant modules in the file are identified. Take into account the current repository's language, frameworks, and dependencies, as well as the existing modules such as helper functions, utility operations and backend services."""

files_to_change_system_prompt = """You are a brilliant and meticulous engineer assigned to plan code changes to solve the following Github issue using the snippets provided from the codebase. You have the utmost care for the plan that you write, so you ensure that all the relevant modules in the file are identified. Take into account the current repository's language, frameworks, and dependencies, as well as the existing modules such as helper functions, utility operations and backend services."""

files_to_change_system_prompt = """\
You are a brilliant and meticulous engineer assigned to plan code changes to solve the following Github issue. You have the utmost care for the plan that you write, so you do not make mistakes and every function and class will be fully implemented. Take into account the current repository's language, frameworks, and dependencies."""


# put more emphasis on modify
# TODO: lots of improvements and cleanup needed here
files_to_change_prompt = """# Task: 
Analyze the provided code snippets, repository, and GitHub issue to understand the requested change. Propose a complete plan for an intern to fully resolve the user's issue, utilizing the relevant code snippets and utility modules provided. Because the intern is unfamiliar with the codebase, provide clear and detailed instructions for updating the code logic.

You are provided with relevent_snippets, which contain code snippets you may need to modify or import and read_only_snippets, which contain code snippets of utility functions, services and type definitions you likely do not need to modify.

Guidelines:
* Always include the full file path (e.g. src/utils/strings/regex_utils.py instead of just strings/regex_utils.py or regex_utils.py) and reference the provided snippets.
* Provide clear, natural language instructions for updating the code logic and specify necessary imports.
* Be specific and direct in your instructions, avoiding vague terms like "identify" or "ensure." Instead, use actionable phrases like "add", "locate" or "change."
* Include relevant type definitions, interfaces, and schemas to provide a clear understanding of the entities and their relationships.
* Avoid using line numbers; instead, reference the locations of the changes using surrounding code or function headers as context.
* When modifying code, provide detailed instructions and the actual code changes required. Write all code changes in the diff format. Do not leave comments or placeholders for the user to fill in.

Please use the following XML format for your response:

# Issue Analysis:
<issue_analysis>
* Identify the root cause of the issue by referencing specific code entities in the relevant files.
* Outline a plan that completely resolves the user's request, referencing provided code snippets, entity names, and necessary files/directories.

List ALL files we should modify to resolve the issue:
- File path 1: Outline of instructions for modifying the file
    - First change to make in the file
    - Second change to make in the file
- File path 2: Outline of instructions for modifying the file
    - First change to make in the file
    - Second change to make in the file
[additional files as needed]

List ALL relevant utility modules from the provided set and specify where they can be used, including:
- Type definitions, interfaces, and schemas
- Helper functions
- Frontend components
- Database services
- API endpoints
[additional relevant modules as needed]

* For each <create> or <modify> section in your plan, explain its purpose and how it contributes to resolving the issue.
* Topologically sort the plan so that changes in low-level modules, such as DB query logic, are made before changes in high-level modules, such as API endpoints.
[additional analysis as needed]
</issue_analysis>

# Plan:
<plan>
<create file="file_path_1" relevant_files="space-separated list of files containing ALL modules to use when creating file_path_1">
* Natural language instructions for creating the new file to solve the issue.
* Reference necessary imports and entity names.
* Include relevant type definitions, interfaces, and schemas.
* Provide the actual code to be added, with detailed explanations.
</create>

[additional creates as needed]

<modify file="file_path_2" relevant_files="space-separated list of files containing ALL modules to use while modifying file_path_2">
* Detailed natural language instructions for modifying the file to solve the issue.
* Reference the locations of the changes using surrounding code or function headers, not line numbers.
* Include relevant type definitions, interfaces, and schemas.
* Each file should be modified at most once. If multiple changes are needed, separate them into different <modify> blocks.
</modify>

[additional modifies as needed]
</plan>

Here's an example of an excellent issue analysis and plan:
<issue_analysis>
The root cause of the issue is that the searchProducts method in the ProductService class (product_service.py) does not properly handle searching for products by category, price range, and keyword simultaneously. It should return products that match all the provided criteria.

To completely resolve the user's request, we need to:

Modify the searchProducts method in product_service.py to:
Add optional parameters for category_id, min_price, max_price, and keyword
Update the database query to filter products based on the provided criteria
Update the search_products endpoint in app.py to:
Extract the new search parameters from the request
Pass the parameters to the searchProducts method
Update the ProductSchema in schemas/product_schema.py to include the category_id field
Relevant files to modify:

src/services/product_service.py - Update searchProducts to handle category, price range, and keyword filtering
src/app.py - Update search_products endpoint to extract new parameters and pass them to searchProducts
src/schemas/product_schema.py - Add category_id field to ProductSchema
Relevant utility modules:
"""

# TODO: Fix relevant files block

files_to_change_prompt = """# Task: 
Critically analyze the provided code snippets, repository, and GitHub issue to understand the requested change. Propose a complete plan for an intern to fully resolve the user's issue, utilizing the relevant code snippets and utility modules provided. Because the intern is unfamiliar with the codebase, provide clear and detailed instructions for updating the code logic.

You are provided with relevent_snippets, which contain code snippets you may need to modify or import and read_only_snippets, which contain code snippets of utility functions, services and type definitions you likely do not need to modify.

Guidelines:
* Always include the full file path and reference the provided snippets.
* Provide clear, natural language instructions for updating the code logic and specify necessary imports.
* Be specific and direct in your instructions, avoiding vague terms like "identify" or "ensure." Instead, use actionable phrases like "add", "locate" or "change."
* Include relevant type definitions, interfaces, and schemas in the relevant_files to provide a clear understanding of the entities and their relationships.
* Avoid using line numbers; instead, reference the locations of the changes using surrounding code or function headers as context.
* Be certain that your plan is complete and covers all the necessary changes to fully resolve the issue.
* Suggest high-quality changes that are completely safe, maintainable, efficient and backwards compatible.
* Divide the task into smaller steps, where each <create> or <modify> section corresponds to one small code block of change. You may have multiple <modify> blocks for the same file.

Please use the following XML format for your response:

# Issue Analysis:
<issue_analysis>
* Identify the root cause of the issue by referencing specific code entities in the relevant files.
* Outline ALL changes that need to occur for the user's request to be resolved, by referencing provided code snippets, entity names, and necessary files/directories.

List ALL files we should modify to resolve the issue:
- File path 1: Outline of instructions for modifying the file
    - First change to make in the file
    - Second change to make in the file
- File path 2: Outline of instructions for modifying the file
    - First change to make in the file
    - Second change to make in the file
[additional files as needed]

List ALL relevant read-only utility modules from the provided set and specify where they can be used. These are not files you need to make changes to but files you need to read while making changes in other files, including:
- Type definitions, interfaces, and schemas
- Helper functions
- Frontend components
- Database services
- API endpoints
[additional relevant modules as needed]

* For each <create> or <modify> section in your plan, explain its purpose and how it contributes to resolving the issue.
[additional analysis as needed]
</issue_analysis>

# Plan:
<plan>
<create file="file_path_1">
* Natural language instructions for creating the new file to solve the issue.
* Reference necessary imports and entity names.
* Include references to relevant type definitions, interfaces, and schemas.
</create>

[additional creates as needed]

<modify file="file_path_2">
* Detailed natural language instructions for modifying the file to solve the issue.
* Reference the locations of the changes using surrounding code or function headers, not line numbers.
* Include references to relevant type definitions, interfaces, and schemas.
* Describe code changes, but do not write the actual code.
* You may modify the same file multiple times. Each <modify> block should contain a single block of code changes from one small section of the file.
</modify>

[additional modifies as needed]
</plan>

<relevant_modules>
[List of all relevant files to use or read while making changes, such as type definitions, interfaces, and schemas, one per line]
</relevant_modules>

Here's an example of an excellent issue analysis and plan:

<issue_analysis>
The root cause of the issue is that the `createPost` method in the `PostService` class (post_service.py) does not validate that the user submitting the post has a non-deleted account. It should check the `user.deleted` property and raise an exception if the user's account is deleted.

To completely resolve the user's request, we need to:
- Modify the `createPost` method in post_service.py to check if the user's account is deleted before creating the post
- Add a new exception class `DeletedAccountError` in exceptions.py to raise when a deleted user tries to create a post
- Update the `create_post` endpoint in app.py to catch the new `DeletedAccountError` and return a 403 error response

Relevant files to modify:
- src/services/post_service.py
  - Import the `User` entity and `DeletedAccountError` 
  - Add validation to check if the user's account is deleted in `createPost`
  - Raise `DeletedAccountError` if the user's account is deleted
- src/exceptions.py
  - Define a new exception class `DeletedAccountError`
- src/app.py
  - Import the new `DeletedAccountError`
  - Catch `DeletedAccountError` in the `create_post` endpoint
  - Return a 403 error response if `DeletedAccountError` is caught

The relevant utility modules are:
- `User` entity (src/entities/user.py) - to check if user's account is deleted 
- `Post` entity (src/entities/post.py) - the entity being created in `createPost`

The <create> and <modify> changes work together to fully handle preventing deleted users from creating posts:
- The new exception class is created first to be used in the other changes
- The service method is updated to validate the user and raise the new exception 
- The API endpoint is updated to catch the new exception and return an appropriate error
</issue_analysis>

<plan>
<create file="src/exceptions.py">
* Define a new exception class called `DeletedAccountError`
* Have it inherit from the base `Exception` class
* Give it a clear error message indicating that a deleted account tried to perform an action
</create>

<modify file="src/services/post_service.py">
At the top of the file:
* Import the `User` entity from `src/entities/user.py`
* Import the `DeletedAccountError` from `src/exceptions.py`

In the `createPost` method of the `PostService` class:
* After getting the `user` by ID, add an if statement to check:
  - If `user.deleted` is True, raise a `DeletedAccountError`
  - Otherwise, continue with creating the post as normal
</modify>

<modify file="src/services/post_service.py">
In the `createPost` method of the `PostService` class:
* After the line that creates the new `post` instance, add:
  - A call to `self.post_repo.save(post)` to save the new post to the database
  - A call to `self.logger.info(f"User {{user.id}} created a new post with id {{post.id}}")` to log the post creation
* Return the newly created `post` instance
</modify>

<modify file="src/app.py">
At the top of the file: 
* Import the `DeletedAccountError` from `src/exceptions.py`

In the `create_post` endpoint:
* Wrap the existing code in a try/except block
* Catch the `DeletedAccountError` in the except block
* If caught, return a JSON response with:
  - A 403 status code
  - An error message like "Cannot create post with a deleted account"
</modify>
</plan>

<relevant_modules>
src/entities/user.py
src/entities/post.py
</relevant_modules>"""

extract_files_to_change_prompt = """\
# Task:
Create a plan that resolves the user's query and ONLY the user's query under "Issue Title" and "Issue Description", providing your response in the below format:
<contextual_request_analysis>
Review each function of each relevant_snippet and analyze the user request to determine if this change should use the refactor or unit test tools.
The refactor tool performs code transformations in a single file without making other logical changes. Determine the function(s) that are too long and should have it's individual parts extracted.
The unit test tool creates or edits unit tests for a given file. Determine all functions that should be unit tested.
</contextual_request_analysis>

<use_tools>
True/False
</use_tools>

If use_tools is True, then generate a plan to use the given tools in this format:
* Make sure destination_module refers to a python module and not a path.

<refactor file="file_path_1" destination_module="destination_module" relevant_files="space-separated list of ALL files relevant for modifying file_path_1">
</refactor>
<test file="file_path_2" source_file="file_path_to_test" relevant_files="space-separated list of ALL files relevant for modifying file_path_2">
* Unit tests for the file_path_to_test, to be written in file_path_2.
* Exact and descriptive instructions for the tests to be created or modified.
...
</test>"""

refactor_files_to_change_prompt = """\
Reference and analyze the snippets, repo, and issue to break down the requested change and propose a plan that addresses the user's request.

Provide a plan to solve the issue, following these rules:
* You may only create new files, extract snippets into functions, relocate functions, or modify existing files.
* Include the full path (e.g. src/main.py and not just main.py), using the repo_tree for reference.
* Be concrete with instructions and do not write "identify x" or "ensure y is done". Simply write "add x" or "change y to z".

You MUST follow the following format:

# Contextual Request Analysis:
<contextual_request_analysis>
* Outline the ideal plan that solves the user request by referencing the snippets and any other necessary files/directories.
* Describe each <create>, <modify>, <extract>, and <relocate> section in the following plan and why it will be needed. Make sure the ordering is correct.
...
</contextual_request_analysis>

# Plan:
<plan>
<create file="file_path_1" relevant_files="space-separated list of ALL files relevant for creating file_path_1">
* Exact instructions for creating the new file needed to solve the issue
* Include references to all files, imports and entity names
...
</create>
...

<extract file="file_path_2" relevant_files="space-separated list of ALL files relevant for modifying file_path_2">
* Extracts lines of code from a function into a new standalone function.
* Only extract lines that reduce the overall nesting or complexity of the code.
...
</extract>
...

<relocate file="file_path_3" new_file_path="new_file_path" handle_references="True">
* This will move a function or variable from file_path_3 into another file while automatically resolving everything. If you use this block do not add other modifications related to imports, references, or function calls.
</relocate>
...

<modify file="file_path_4" relevant_files="space-separated list of ALL files relevant for modifying file_path_4">
* Modifies files by making less than 30-line changes. Be exact and mention references to all files, imports and entity names.
* Use detailed, natural language instructions on what to modify regarding business logic, and reference files to import.
* Be concrete with instructions and do not write "identify x" or "ensure y is done". Simply write "add x" or "change y to z".
* You may modify the same file multiple times.
...
</modify>
...

</plan>"""

sandbox_files_to_change_prompt = """\
Analyze the snippets, repo, and issue to break down the requested problem or feature. Then propose a high-quality plan that completely fixes the CI/CD run.

Provide a list of ALL of the files we should modify, abiding by the following:
* You may only create and modify files.
* Including the FULL path, e.g. src/main.py and not just main.py, using the repo_tree for reference.
* Use detailed, natural language instructions on what to modify regarding business logic, and reference files to import.
* Be concrete with instructions and do not write "check for x" or "ensure y is done". Simply write "add x" or "change y to z".
* If the tests fail you should typically fix the tests and not the source code.

You MUST follow the following format with the final output in XML tags:

<analysis_and_plan>
Whether the change was caused by the user's change or not. If not, then leave the plan empty.
Otherwise, determine why the CI/CD run failed and the root cause. Determine the MINIMAL amount of changes to fix, with reference to entities, in the following format:

<minimal_changes>
* Change x: file to make the change
* Change y: file to make the change
...
</minimal_changes>
</analysis_and_plan>

<plan>
<create file="file_path_1" relevant_files="space-separated list of ALL files relevant for creating file_path_1">
Outline of additions in concise natural language of what needs to be implemented in this file, referencing to external and imported libraries and business logic.
</create>

<modify file="file_path_2" relevant_files="space-separated list of ALL files relevant for modifying file_path_2">
Outline of modifications in natural language (no code), referencing entities, and what type of patterns to look for, such as all occurrences of a variable or function call.
Do not make this XML block if no changes are needed.
</modify>
...

</plan>"""

subissues_prompt = """
Think step-by-step to break down the requested problem into sub-issues each of equally sized non-trivial changes. The sub-issue should be a small, self-contained, and independent part of the problem, and should partition the files to be changed.
You MUST follow the following format with the final output in XML tags:

Root cause:
Identify the root cause of this issue and a minimum plan to address this issue concisely in two sentences.


Step-by-step thoughts with explanations:
* Concise imperative thoughts
* No conjunctions
...

<plan>
<issue title="title_1">
* In file_path_1, do a
* In file_path_1, do b
...
* In file_path_2, do c
* In file_path_2, do d
...
</issue>

<issue title="title_2">
* In file_path_1, do a
* In file_path_1, do b
...
* In file_path_2, do c
* In file_path_2, do d
...
</issue>

...
</plan>"""

create_file_prompt = """You are creating a file of code as part of a PR to solve the GitHub user's request. You will follow the request under "# Request" and respond based on the format under "# Format".

# Request

file_name: "{filename}"

{instructions}

# Format

You MUST respond in the following XML format:

<contextual_request_analysis>
Concisely analyze the request and list step-by-step thoughts on what to create in each section, with low-level, detailed references to functions, variables, and imports to create, and what each function does. Be as explicit and specific as possible.
Maximize information density in this section.
</contextual_request_analysis>

<new_file>
The contents of the new file. NEVER write comments. All functions and classes will be fully implemented.
When writing unit tests, they will be complete, extensive, and cover ALL edge cases. You will make up data for unit tests. Create mocks when necessary.
</new_file>

Commit message: "feat/fix: the commit message\"""".strip()

"""
Reply in the format below.
* You MUST use the new_file XML tags
* DO NOT write ``` anywhere, unless it's markdown
* DO NOT write "pass" or "Rest of code"
* Do not literally write "{{new_file}}".
* Format:
"""

chunking_prompt = """
We are handling this file in chunks. You have been provided a section of the code.
Any lines that you do not see will be handled, so trust that the imports are managed and any other issues are taken care of.
If you see code that should be modified, please modify it. The changes may not need to be in this chunk, do not make any changes."""

modify_file_hallucination_prompt = [
    {
        "content": """File Name: (non-existent example)
<old_file>
example = True
if example:
    x = 1 # comment
    print("hello")
    x = 2

class Example:
    foo: int = 1

    def func():
        a = 3

</old_file>

---

Code Planning:
Step-by-step thoughts with explanations:
* Thought 1
* Thought 2
...

Commit message: "feat/fix: the commit message"

Detailed plan of modifications:
* Modification 1
* Modification 2
...

Code Generation:

```
Generate a diff based on the given plan using the search and replace pairs in the format below.
* Always prefer the least amount of changes possible, but ensure the solution is complete
* Prefer multiple small changes over a single large change.
* NEVER write ellipses anywhere in the diffs. Simply write two diff hunks: one for the beginning and another for the end.
* Always add lines before and after. The ORIGINAL section should be at least 5 lines long.

The format is as follows:

<<<< ORIGINAL
line_before
old_code
line_after
====
line_before
new_code
line_after
>>>> UPDATED
```

Commit message: "the commit message"

Request: "Change hello to goodbye and change 3 to 4". Limit your changes to the request.

Instructions:
1. Complete the Code Planning step
2. Complete the Code Generation step""",
        "role": "user",
        "key": "modify_file_hallucination",
    },
    {
        "content": """Code Planning:
Step-by-step thoughts with explanations:
* We need to print "goodbye" instead of "hello".
* We need to update the value of the variable a from 3 to 4.

Detailed plan of modifications:
* Change the output of the print statement from "hello" to "goodbye" as an example modification.
* I will update the value of a from 3 to 4.

Code Generation:
```
<<<< ORIGINAL
example = True
if example:
    x = 1 # comment
    print("hello")
    x = 2
====
example = True
if example:
    x = 1 # comment
    print("goodbye")
    x = 2
>>>> UPDATED

<<<< ORIGINAL
class Example:
    foo: int = 1

    def func():
        a = 3
====
class Example:
    foo: int = 1

    def func():
        a = 4
>>>> UPDATED
```

Commit message: "Changed goodbye to hello and 3 to 4"\
""",
        "role": "assistant",
        "key": "modify_file_hallucination",
    },
]

# TODO: IMPORTANT: THIS DEPENDS ON THE ABOVE PROMPT, modify_file_hallucination_prompt
modify_file_prompt_3 = """\
File Name: {filename}
<old_file>
{code}
</old_file>

---

User's request:
{instructions}

Limit your changes to the request.

Instructions:
Complete the Code Planning step and Code Modification step.
Remember to NOT write ellipses, code things out in full, and use multiple small hunks.\
"""

modify_recreate_file_prompt_3 = """\
File Name: {filename}
<old_file>
{code}
</old_file>

---

User's request:
{instructions}

Limit your changes to the request.

Format:
```
<new_file>
{{new file content}}
</new_file>
```

Instructions:
1. Complete the Code Planning step
2. Complete the Code Modification step, remembering to NOT write ellipses, write complete functions, and use multiple small hunks where possible."""

modify_file_system_message = """\
You are a brilliant and meticulous engineer assigned to write code for the file to address a Github issue. When you write code, the code works on the first try and is syntactically perfect and complete. You have the utmost care for your code, so you do not make mistakes and every function and class will be fully implemented. Take into account the current repository's language, frameworks, and dependencies. You always follow up each code planning session with a code modification.

When you modify code:
* Always prefer the least amount of changes possible, but ensure the solution is complete.
* Prefer multiple small changes over a single large change.
* Do not edit the same parts multiple times.
* Make sure to add additional lines before and after the original and updated code to disambiguate code when replacing repetitive sections.
* NEVER write ellipses anywhere in the diffs. Simply write two diff hunks: one for the beginning and another for the end.

Respond in the following format. Both the Code Planning and Code Modification steps are required.

### Format ###

## Code Planning:

Thoughts and detailed plan:
1.
2.
3.
...

Commit message: "feat/fix: the commit message"

## Code Modification:

Generated diff hunks based on the given plan using the search and replace pairs in the format below.
```
The first hunk's description

<<<< ORIGINAL
{exact copy of lines you would like to change}
====
{updated lines}
>>>> UPDATED

The second hunk's description

<<<< ORIGINAL
second line before
first line before
old code
first line after
second line after
====
second line before
first line before
new code
first line after
second line after
>>>> UPDATED
```"""

RECREATE_LINE_LENGTH = -1

modify_file_prompt_4 = """\
File Name: {filename}

<file>
{code}
</file>

---

Modify the file by responding in the following format:

Code Planning:

Step-by-step thoughts with explanations:
* Thought 1
* Thought 2
...

Detailed plan of modifications:
* Replace x with y
* Add a foo method to bar
...

Code Modification:

```
Generate a diff based on the given instructions using the search and replace pairs in the following format:

<<<< ORIGINAL
second line before
first line before
old code
first line after
second line after
====
second line before
first line before
new code
first line after
second line after
>>>> UPDATED
```

Commit message: "the commit message"

The user's request is:
{instructions}

Instructions:
1. Complete the Code Planning step
2. Complete the Code Modification step
"""

rewrite_file_system_prompt = "You are a brilliant and meticulous engineer assigned to write code for the file to address a Github issue. When you write code, the code works on the first try and is syntactically perfect and complete. You have the utmost care for your code, so you do not make mistakes and every function and class will be fully implemented. Take into account the current repository's language, frameworks, and dependencies."

rewrite_file_prompt = """\
File Name: {filename}
<old_file>
{code}
</old_file>

---

User's request:
{instructions}

Limit your changes to the request.

Rewrite the following section from the old_file to handle this request.

<section>

{section}

</section>

Think step-by-step on what to modify, then wrap the final answer in the brackets <section></section> XML tags. Only rewrite the section and do not close hanging parentheses and tags.\
"""

sandbox_code_repair_modify_prompt_2 = """
File Name: {filename}

<file>
{code}
</file>

---

Above is the code that was written by an inexperienced programmer, and contain errors such as syntax errors, linting erors and type-checking errors. The CI pipeline returned the following logs:

stdout:
```
{stdout}
```

stderr
```
{stderr}
```

Respond in the following format:

Code Planning

Determine the following in code planning:
1. Are there any syntax errors? Look through the file to find all syntax errors.
2. Are there basic linting errors, like undefined variables, undefined members or type errors?
3. Are there incorrect imports and exports?
4. Are there any other errors not listed above?

Determine whether changes are necessary based on the errors (ignore warnings).

Code Modification:

Generate a diff based on the given plan using the search and replace pairs in the format below.
* Always prefer the least amount of changes possible, but ensure the solution is complete
* Prefer multiple small changes over a single large change.
* NEVER write ellipses anywhere in the diffs. Simply write two diff hunks: one for the beginning and another for the end.
* DO NOT modify the same section multiple times.
* Always add lines before and after. The ORIGINAL section should be at least 5 lines long.
* Restrict the changes to fixing the errors from the logs.

The format is as follows:

```
<<<< ORIGINAL
second line before
first line before
old code of first hunk
first line after
second line after
====
second line before
first line before
new code of first hunk
first line after
second line after
>>>> UPDATED

<<<< ORIGINAL
second line before
first line before
old code of second hunk
first line after
second line after
====
second line before
first line before
new code of second hunk
first line after
second line after
>>>> UPDATED
```

Commit message: "the commit message"

Instructions:
1. Complete the Code Planning step
2. Complete the Code Modification step
"""

pr_code_prompt = ""  # TODO: deprecate this

pull_request_prompt = """Now, create a PR for your changes. Be concise but cover all of the changes that were made.
For the pr_content, add two sections, description and summary.
Use GitHub markdown in the following format:

pr_title = "..."
branch = "..."
pr_content = \"\"\"
...
...
\"\"\""""

summarize_system_prompt = """
You are an engineer assigned to helping summarize code instructions and code changes.
"""

user_file_change_summarize_prompt = """
Summarize the given instructions for making changes in a pull request.
Code Instructions:
{message_content}
"""

assistant_file_change_summarize_prompt = """
Please summarize the following file using the file stubs.
Be sure to repeat each method signature and docstring. You may also add additional comments to the docstring.
Do not repeat the code in the file stubs.
Code Changes:
{message_content}
"""

code_repair_check_system_prompt = """\
You are a genius trained for validating code.
You will be given two pieces of code marked by xml tags. The code inside <diff></diff> is the changes applied to create user_code, and the code inside <user_code></user_code> is the final product.
Our goal is to validate if the final code is valid. This means there are no undefined variables, no syntax errors, has no unimplemented functions (e.g. pass's, comments saying "rest of code") and the code runs.
"""

code_repair_check_prompt = """\
This is the diff that was applied to create user_code. Only make changes to code in user_code if the code was affected by the diff.

This is the user_code.
<user_code>
{user_code}
</user_code>

Reply in the following format:

Step-by-step thoughts with explanations:
1. No syntax errors: True/False
2. No undefined variables: True/False
3. No unimplemented functions: True/False
4. Code runs: True/False

<valid>True</valid> or <valid>False</valid>
"""

code_repair_system_prompt = """\
You are a genius trained for code stitching.
You will be given two pieces of code marked by xml tags. The code inside <diff></diff> is the changes applied to create user_code, and the code inside <user_code></user_code> is the final product. The intention was to implement a change described as {feature}.
Our goal is to return a working version of user_code that follows {feature}. We should follow the instructions and make as few edits as possible.
"""

code_repair_prompt = """\
This is the diff that was applied to create user_code. Only make changes to code in user_code if the code was affected by the diff.

This is the user_code.
<user_code>
{user_code}
</user_code>

Instructions:
* Do not modify comments, docstrings, or whitespace.

The only operations you may perform are:
1. Indenting or dedenting code in user_code. This code MUST be code that was modified by the diff.
2. Adding or deduplicating code in user_code. This code MUST be code that was modified by the diff.

Return the working user_code without xml tags. All of the text you return will be placed in the file.
"""

doc_query_rewriter_system_prompt = """\
You must rewrite the user's github issue to leverage the docs. In this case we want to look at {package}. It's used for: {description}. Using the github issue, write a search query that searches for the potential answer using the documentation. This query will be sent to a documentation search engine with vector and lexical based indexing. Make this query contain keywords relevant to the {package} documentation.
"""

doc_query_rewriter_prompt = """\
This is the issue:
{issue}

Write a comprehensive search query for the answer.
"""

should_edit_code_system_prompt = """\
We are processing a large file and trying to make code changes to it.
The file is definitely relevant, but the section we observe may not be relevant.
Your job is to determine whether the instructions are referring to the given section of the file.
"""

should_edit_code_prompt = """\
Here are the instructions to change the code in the file:
{problem_description}
Here is the code snippet from the file:
{code_snippet}

To determine whether the instructions are referring to this section of the file, respond in the following format:
1. Step-by-step thoughts with explanations:
* Thought 1 - Explanation 1
* Thought 2 - Explanation 2
...
2. Planning:
* Is the code relevant?
* If so, what is the relevant part of the code?
* If not, what is the reason?

3. In the last line of your response, write either <relevant>True</relevant> or <relevant>False</relevant>.
"""

slow_mode_system_prompt = """You are a brilliant and meticulous software architect. Your job is to take in the user's GitHub issue and the relevant context from their repository to:
1. Gather more code snippets using a code search engine.
2. Expand upon the plan to address the issue."""

generate_plan_and_queries_prompt = """Think step-by-step to break down the requested problem or feature, and write up to three queries for the code search engine. These queries should find relevant code that is not already mentioned in the existing snippets. They should all mention different files and subtasks of the initial issue, avoid duplicates.
Then add more instructions that build on the user's instructions. These instructions should help plan how to solve the issue.
* The code search engine is based on semantic similarity. Ask questions that involve code snippets, function references, or mention relevant file paths.
* The user's instructions should be treated as the source of truth, but sometimes the user will not mention the entire context. In that case, you should add the missing context.
* Gather files that are relevant, including dependencies and similar files in the codebase. For example, if the user asked to write tests, look at similar tests.

You MUST follow the following format delimited with XML tags:

Step-by-step thoughts with explanations:
* Thought 1 - Explanation 1
* Thought 2 - Explanation 2
...

<queries>
* query 1
* query 2
* query 3
</queries>

<additional_instructions>
* additional instructions to be appended to the user's instructions
</additional_instructions>
"""

external_search_system_prompt = (
    "You are an expert at summarizing content from pages that are relevant to a query."
    " You will be given a page and asked to summarize it."
)

external_search_prompt = """\
Here is the page metadata:
```
{page_metadata}
```

The user is attempting to solve the following problem:
\'\'\'
{problem}
\'\'\'

Provide a summary of the page relevant to the problem, including all code snippets.
"""


docs_qa_system_prompt = """You are an expert at summarizing documentation for programming-related to help the user solve the problem. You will be given a question and relevant snippets of documentation, and be asked to provide a summary of relevant snippets for solving the problem."""
docs_qa_user_prompt = """Here are the relevant documentation snippets:
{snippets}

The user is attempting to solve the following problem:
{problem}
"""

linting_new_file_prompt = """
<new_file>
{code}
</new_file>
"""

linting_modify_prompt = """Code Planning:
Step-by-step thoughts with explanations:
* Thought 1
* Thought 2
...

Detailed plan of modifications:
* Modification 1
* Modification 2
...

Code Generation:

```
Generate a diff based on the given plan using the search and replace pairs in the format below.
* Always prefer the least amount of changes possible
* Prefer many small edits over few large edits
* Always add lines before and after. The ORIGINAL section should be at least 5 lines long.

The linter returned the following logs:
<linter_logs>
{logs}
</linter_logs>

Modify your created file. Note if no changes are needed.

---

The format is as follows:

<<<< ORIGINAL
====
>>>> UPDATED

Instructions:
1. Complete the Code Planning step
2. Complete the Code Generation step"""

summarize_snippet_system_prompt = """You are a technical writer. Summarize code for an engineer. Be concise but detailed. Mention all entities implicitly. Never say "the code does x", just say "x". Keep it within 30 lines.

E.g: If `issue_comment` is None, set `issue_comment` to `current_issue.create_comment(first_comment)`, otherwise edit comment via `issue_comment.edit(first_comment)`"""

summarize_snippet_prompt = """# Code
```
{code}
```

# Repo Metadata
{metadata}

# Issue
{issue}

# Instructions
Losslessly summarize the code in a ordered list for an engineer to search for relevant code to solve the above GitHub issue."""


use_chunking_message = """\
Respond with a list of the MINIMUM sections that should be modified. To insert code after a function, fetch the last few lines of the function."""

dont_use_chunking_message = """\
Respond with a list of the MINIMUM sections that should be modified. To insert code after a function, fetch the last few lines of the function."""



from collections import defaultdict
import traceback
from time import time

import cohere
from loguru import logger
from tqdm import tqdm

from sweepai.config.client import SweepConfig, get_blocked_dirs
from sweepai.config.server import COHERE_API_KEY
from sweepai.core.context_pruning import RepoContextManager, get_relevant_context, integrate_graph_retrieval
from sweepai.core.entities import Snippet
from sweepai.core.lexical_search import (
    compute_vector_search_scores,
    prepare_lexical_search_index,
    search_index,
)
from sweepai.logn.cache import file_cache
from sweepai.utils.chat_logger import discord_log_error
from sweepai.utils.event_logger import posthog
from sweepai.utils.github_utils import ClonedRepo
from sweepai.utils.multi_query import generate_multi_queries
from sweepai.utils.openai_listwise_reranker import listwise_rerank_snippets
from sweepai.utils.progress import TicketProgress

"""
Input queries are in natural language so both lexical search 
and vector search have a heavy bias towards natural language
files such as tests, docs and localization files. Therefore,
we add adjustment scores to compensate for this bias.
"""

prefix_adjustment = {
    ".": 0.5,
    "doc": 0.3,
    "example": 0.7,
}

suffix_adjustment = {
    ".cfg": 0.8,
    ".ini": 0.8,
    ".txt": 0.8,
    ".rst": 0.8,
    ".md": 0.8,
    ".html": 0.8,
    ".po": 0.5,
    ".json": 0.8,
    ".toml": 0.8,
    ".yaml": 0.8,
    ".yml": 0.8,
    ".1": 0.5, # man pages
    ".spec.ts": 0.6,
    ".spec.js": 0.6,
    ".generated.ts": 0.5,
    ".generated.graphql": 0.5,
    ".generated.js": 0.5,
    "ChangeLog": 0.5,
}

substring_adjustment = {
    "tests/": 0.5,
    "test_": 0.5,
    "_test": 0.5,
    "egg-info": 0.5,
    "LICENSE": 0.5,
}

def apply_adjustment_score(
    snippet: str,
    old_score: float,
):
    snippet_score = old_score
    file_path, *_ = snippet.split(":")
    file_path = file_path.lower()
    for prefix, adjustment in prefix_adjustment.items():
        if file_path.startswith(prefix):
            snippet_score *= adjustment
            break
    for suffix, adjustment in suffix_adjustment.items():
        if file_path.endswith(suffix):
            snippet_score *= adjustment
            break
    for substring, adjustment in substring_adjustment.items():
        if substring in file_path:
            snippet_score *= adjustment
            break
    # Penalize numbers as they are usually examples of:
    # 1. Test files (e.g. test_utils_3*.py)
    # 2. Generated files (from builds or snapshot tests)
    # 3. Versioned files (e.g. v1.2.3)
    # 4. Migration files (e.g. 2022_01_01_*.sql)
    base_file_name = file_path.split("/")[-1]
    num_numbers = sum(c.isdigit() for c in base_file_name)
    snippet_score *= (1 - 1 / len(base_file_name)) ** num_numbers
    return snippet_score

NUM_SNIPPETS_TO_RERANK = 100

@file_cache()
def multi_get_top_k_snippets(
    cloned_repo: ClonedRepo,
    queries: list[str],
    ticket_progress: TicketProgress | None = None,
    k: int = 15,
):
    """
    Handles multiple queries at once now. Makes the vector search faster.
    """
    sweep_config: SweepConfig = SweepConfig()
    blocked_dirs = get_blocked_dirs(cloned_repo.repo)
    sweep_config.exclude_dirs += blocked_dirs
    _, snippets, lexical_index = prepare_lexical_search_index(
        cloned_repo.cached_dir,
        sweep_config,
        ticket_progress,
        ref_name=f"{str(cloned_repo.git_repo.head.commit.hexsha)}",
    )
    if ticket_progress:
        ticket_progress.search_progress.indexing_progress = (
            ticket_progress.search_progress.indexing_total
        )
        ticket_progress.save()

    for snippet in snippets:
        snippet.file_path = snippet.file_path[len(cloned_repo.cached_dir) + 1 :]
    # We can mget the lexical search scores for all queries at once
    # But it's not that slow anyways
    content_to_lexical_score_list = [search_index(query, lexical_index) for query in queries]
    files_to_scores_list = compute_vector_search_scores(queries, snippets)

    for i, query in enumerate(queries):
        for snippet in tqdm(snippets):
            vector_score = files_to_scores_list[i].get(snippet.denotation, 0.04)
            snippet_score = 0.02
            if snippet.denotation in content_to_lexical_score_list[i]:
                # roughly fine tuned vector score weight based on average score from search_eval.py on 10 test cases Feb. 13, 2024
                snippet_score = content_to_lexical_score_list[i][snippet.denotation] + (
                    vector_score * 3.5
                )
                content_to_lexical_score_list[i][snippet.denotation] = snippet_score
            else:
                content_to_lexical_score_list[i][snippet.denotation] = snippet_score * vector_score
            content_to_lexical_score_list[i][snippet.denotation] = apply_adjustment_score(
                snippet.denotation, content_to_lexical_score_list[i][snippet.denotation]
            )
    
    ranked_snippets_list = [
        sorted(
            snippets,
            key=lambda snippet: content_to_lexical_score[snippet.denotation],
            reverse=True,
        )[:k] for content_to_lexical_score in content_to_lexical_score_list
    ]
    return ranked_snippets_list, snippets, content_to_lexical_score_list

@file_cache()
def get_top_k_snippets(
    cloned_repo: ClonedRepo,
    query: str,
    ticket_progress: TicketProgress | None = None,
    k: int = 15,
):
    ranked_snippets_list, snippets, content_to_lexical_score_list = multi_get_top_k_snippets(
        cloned_repo, [query], ticket_progress, k
    )
    return ranked_snippets_list[0], snippets, content_to_lexical_score_list[0]

@file_cache()
def cohere_rerank_call(
    query: str,
    documents: list[str],
    model='rerank-english-v3.0',
    **kwargs,
):
    # Cohere API call with caching
    co = cohere.Client(COHERE_API_KEY)
    return co.rerank(
        model=model,
        query=query,
        documents=documents,
        **kwargs
    )

def get_pointwise_reranked_snippet_scores(
    query: str,
    snippets: list[Snippet],
    snippet_scores: dict[str, float],
):
    """
    Ranks 1-5 snippets are frozen. They're just passed into Cohere since it helps with reranking. We multiply the scores by 1_000 to make them more significant.
    Ranks 6-100 are reranked using Cohere. Then we divide the scores by 1_000 to make them comparable to the original scores.
    """

    if not COHERE_API_KEY:
        return snippet_scores

    sorted_snippets = sorted(
        snippets,
        key=lambda snippet: snippet_scores[snippet.denotation],
        reverse=True,
    )

    NUM_SNIPPETS_TO_KEEP = 5
    NUM_SNIPPETS_TO_RERANK = 100

    response = cohere_rerank_call(
        model='rerank-english-v3.0',
        query=query,
        documents=[snippet.xml for snippet in sorted_snippets[:NUM_SNIPPETS_TO_RERANK]],
        max_chunks_per_doc=900 // NUM_SNIPPETS_TO_RERANK,
    )

    new_snippet_scores = {k: v / 1000 for k, v in snippet_scores.items()}

    for document in response.results:
        new_snippet_scores[sorted_snippets[document.index].denotation] = apply_adjustment_score(
            sorted_snippets[document.index].denotation,
            document.relevance_score,
        )

    for snippet in sorted_snippets[:NUM_SNIPPETS_TO_KEEP]:
        new_snippet_scores[snippet.denotation] = snippet_scores[snippet.denotation] * 1_000
    
    return new_snippet_scores

def multi_prep_snippets(
    cloned_repo: ClonedRepo,
    queries: list[str],
    ticket_progress: TicketProgress | None = None,
    k: int = 15,
    skip_reranking: bool = False, # This is only for pointwise reranking
    skip_pointwise_reranking: bool = False,
) -> RepoContextManager:
    """
    Assume 0th index is the main query.
    """
    rank_fusion_offset = 0
    if len(queries) > 1:
        logger.info("Using multi query...")
        ranked_snippets_list, snippets, content_to_lexical_score_list = multi_get_top_k_snippets(
            cloned_repo, queries, ticket_progress, k * 3 # k * 3 to have enough snippets to rerank
        )
        # Use RRF to rerank snippets
        content_to_lexical_score = defaultdict(float)
        for i, ordered_snippets in enumerate(ranked_snippets_list):
            for j, snippet in enumerate(ordered_snippets):
                content_to_lexical_score[snippet.denotation] += content_to_lexical_score_list[i][snippet.denotation] * (1 / 2 ** (rank_fusion_offset + j))
        if not skip_pointwise_reranking:
            content_to_lexical_score = get_pointwise_reranked_snippet_scores(
                queries[0], snippets, content_to_lexical_score
            )
        ranked_snippets = sorted(
            snippets,
            key=lambda snippet: content_to_lexical_score[snippet.denotation],
            reverse=True,
        )[:k]
    else:
        ranked_snippets, snippets, content_to_lexical_score = get_top_k_snippets(
            cloned_repo, queries[0], ticket_progress, k
        )
        if not skip_pointwise_reranking:
            content_to_lexical_score = get_pointwise_reranked_snippet_scores(
                queries[0], snippets, content_to_lexical_score
            )
        ranked_snippets = sorted(
            snippets,
            key=lambda snippet: content_to_lexical_score[snippet.denotation],
            reverse=True,
        )[:k]
    if ticket_progress:
        ticket_progress.search_progress.retrieved_snippets = ranked_snippets
        ticket_progress.save()
    # you can use snippet.denotation and snippet.get_snippet()
    if not skip_reranking:
        ranked_snippets[:NUM_SNIPPETS_TO_RERANK] = listwise_rerank_snippets(queries[0], ranked_snippets[:NUM_SNIPPETS_TO_RERANK])
    snippet_paths = [snippet.file_path for snippet in ranked_snippets]
    prefixes = []
    for snippet_path in snippet_paths:
        snippet_depth = len(snippet_path.split("/"))
        for idx in range(snippet_depth):  # heuristic
            if idx > snippet_depth // 2:
                prefixes.append("/".join(snippet_path.split("/")[:idx]) + "/")
        prefixes.append(snippet_path)
    _, dir_obj = cloned_repo.list_directory_tree(
        included_directories=list(set(prefixes)),
        included_files=list(set(snippet_paths)),
    )
    repo_context_manager = RepoContextManager(
        dir_obj=dir_obj,
        current_top_tree=str(dir_obj),
        current_top_snippets=ranked_snippets,
        snippets=snippets,
        snippet_scores=content_to_lexical_score,
        cloned_repo=cloned_repo,
    )
    return repo_context_manager

def prep_snippets(
    cloned_repo: ClonedRepo,
    query: str,
    ticket_progress: TicketProgress | None = None,
    k: int = 15,
    skip_reranking: bool = False,
    use_multi_query: bool = True,
) -> RepoContextManager:
    if use_multi_query:
        queries = [query, *generate_multi_queries(query)]
    else:
        queries = [query]
    return multi_prep_snippets(
        cloned_repo, queries, ticket_progress, k, skip_reranking
    )

def fetch_relevant_files(
    cloned_repo,
    title,
    summary,
    replies_text,
    username,
    metadata,
    on_ticket_start_time,
    tracking_id,
    is_paying_user,
    is_consumer_tier,
    issue_url,
    chat_logger,
    ticket_progress: TicketProgress,
):
    logger.info("Fetching relevant files...")
    try:
        search_query = (title + summary + replies_text).strip("\n")
        replies_text = f"\n{replies_text}" if replies_text else ""
        formatted_query = (f"{title.strip()}\n{summary.strip()}" + replies_text).strip(
            "\n"
        )
        repo_context_manager = prep_snippets(cloned_repo, search_query, ticket_progress)

        repo_context_manager, import_graph = integrate_graph_retrieval(search_query, repo_context_manager)

        ticket_progress.search_progress.repo_tree = str(repo_context_manager.dir_obj)
        ticket_progress.save()
        repo_context_manager = get_relevant_context(
            formatted_query,
            repo_context_manager,
            ticket_progress,
            chat_logger=chat_logger,
            import_graph=import_graph,
        )
        snippets = repo_context_manager.current_top_snippets
        ticket_progress.search_progress.repo_tree = str(repo_context_manager.dir_obj)
        ticket_progress.search_progress.final_snippets = snippets
        ticket_progress.save()

        tree = str(repo_context_manager.dir_obj)
        dir_obj = repo_context_manager.dir_obj
    except Exception as e:
        trace = traceback.format_exc()
        logger.exception(f"{trace} (tracking ID: `{tracking_id}`)")
        log_error(
            is_paying_user,
            is_consumer_tier,
            username,
            issue_url,
            "File Fetch",
            str(e) + "\n" + traceback.format_exc(),
            priority=1,
        )
        posthog.capture(
            username,
            "failed",
            properties={
                **metadata,
                "error": str(e),
                "duration": time() - on_ticket_start_time,
            },
        )
        raise e
    return snippets, tree, dir_obj, repo_context_manager


SLOW_MODE = False
SLOW_MODE = True


def log_error(
    is_paying_user,
    is_trial_user,
    username,
    issue_url,
    error_type,
    exception,
    priority=0,
):
    if is_paying_user or is_trial_user:
        if priority == 1:
            priority = 0
        elif priority == 2:
            priority = 1

    prefix = ""
    if is_trial_user:
        prefix = " (TRIAL)"
    if is_paying_user:
        prefix = " (PRO)"

    content = (
        f"**{error_type} Error**{prefix}\n{username}:"
        f" {issue_url}\n```{exception}```"
    )
    discord_log_error(content, priority=2)


def center(text: str) -> str:
    return f"<div align='center'>{text}</div>"


def fire_and_forget_wrapper(call):
    """
    This decorator is used to run a function in a separate thread.
    It does not return anything and does not wait for the function to finish.
    It fails silently.
    """

    def wrapper(*args, **kwargs):
        try:
            return call(*args, **kwargs)
        except Exception:
            pass
        # def run_in_thread(call, *a, **kw):
        #     try:
        #         call(*a, **kw)
        #     except:
        #         pass

        # thread = Thread(target=run_in_thread, args=(call,) + args, kwargs=kwargs)
        # thread.start()

    return wrapper

if __name__ == "__main__":
    from sweepai.utils.github_utils import MockClonedRepo

    cloned_repo = MockClonedRepo(
        _repo_dir="/tmp/sweep",
        repo_full_name="sweepai/sweep",
    )
    cloned_repo = MockClonedRepo(
        _repo_dir="/tmp/pulse-alp",
        repo_full_name="trilogy-group/pulse-alp",
    )
    rcm = prep_snippets(
        cloned_repo,
        # "I am trying to set up payment processing in my app using Stripe, but I keep getting a 400 error when I try to create a payment intent. I have checked the API key and the request body, but I can't figure out what's wrong. Here is the error message I'm getting: 'Invalid request: request parameters are invalid'. I have attached the relevant code snippets below. Can you help me find the part of the code that is causing this error?",
        "Where can I find the section that checks if assembly line workers are active or disabled?",
        use_multi_query=False,
        skip_reranking=True
    )
    breakpoint()




# ensure that all additional_messages are 32768 characters at most, if not split them
import textwrap
from sweepai.core.entities import Message

MAX_CHARS = 32000

def ensure_additional_messages_length(additional_messages: list[Message]) -> list[Message]:
    for i, additional_message in enumerate(additional_messages):
        if len(additional_message.content) > MAX_CHARS:
            wrapper = textwrap.TextWrapper(width=MAX_CHARS, replace_whitespace=False)
            new_messages = wrapper.wrap(additional_message.content)
            # replace the original message with the broken up messages
            for j, new_message in enumerate(new_messages):
                if j == 0:
                    additional_messages[i] = Message(
                        role=additional_message.role,
                        content=new_message,
                    )
                else:
                    additional_messages.insert(
                        i + j,
                        Message(
                            role=additional_message.role,
                            content=new_message,
                        ),
                    )
    return additional_messages


import json
import os
import unittest
from unittest.mock import mock_open, patch

from sweepai.cli import load_config, run


class TestConfigLoading(unittest.TestCase):
    @patch.dict(os.environ, {}, clear=True)
    @patch("os.path.exists", return_value=True)
    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data=json.dumps(
            {"GITHUB_PAT": "test_github_pat", "OPENAI_API_KEY": "test_openai_key"}
        ),
    )
    def test_load_config(self, mock_file, mock_exists):
        with self.assertRaises(KeyError):
            os.environ["GITHUB_PAT"]
        with self.assertRaises(KeyError):
            os.environ["OPENAI_API_KEY"]
        load_config()
        self.assertEqual(os.environ["GITHUB_PAT"], "test_github_pat")
        self.assertEqual(os.environ["OPENAI_API_KEY"], "test_openai_key")


class TestCLIRunIsssue(unittest.TestCase):
    @patch("os.path.exists", return_value=False)
    def test_run_issue_no_config(self, mock_exists):
        with self.assertRaises(ValueError):
            run("https://github.com/sweepai/e2e/issues/1")


if __name__ == "__main__":
    unittest.main()



import unittest

from sweepai.utils.comment_utils import check_comments_presence


class TestCommentUtils(unittest.TestCase):
    def test_check_comments_presence(self):
        # Test Python file with a comment
        self.assertEqual(
            check_comments_presence("test.py", "# This is a comment"), True
        )

        # Test Python file without a comment
        self.assertEqual(
            check_comments_presence("test.py", 'print("Hello, World!")'), False
        )

        # Test JavaScript file with a comment
        self.assertEqual(
            check_comments_presence("test.js", "// This is a comment"), True
        )

        # Test JavaScript file without a comment
        self.assertEqual(
            check_comments_presence("test.js", 'console.log("Hello, World!");'), False
        )

        # Test unsupported file type with a comment
        self.assertEqual(
            check_comments_presence("test.txt", "# This is a comment"), False
        )

        # Test unsupported file type without a comment
        self.assertEqual(check_comments_presence("test.txt", "Hello, World!"), False)


if __name__ == "__main__":
    unittest.main()
